/****************************************************************************/
/*! \file		chsim.c
	\brief		Phone line channel simulator
	\author		Iliya Voronov, http://powerdsp.narod.ru
	\version	0.0
     
	Phone line model for modem development.
	Model simulates following channel distortions:
	1. delay and attenuation
	2. linear distortion
	3. currier offset
	4. intermodulation distortion
	5. additive wight Gaussian noise
	7. near and far echo
	8. baud rate offset

	Model process samples frame by frame, sample format is int16, 
	max input value is -8192 to 8191. Frame size is CHSIM_FR_SZ 
	at Modem0 side and CHSIM_FR_SZ +/- 1 at Modem1 side.

	Model should be initialized with CHSIM_init, 
	then model parameters should be configured by calling CHSIM_parseStr
	for each string from configuration file,
	then model can be used to process samples with CHSIM_proc.

	Channel parameters are defined in file chsim_param.c
	Line parsing commands are define in chsim_parse.c

	TODO: Implement resampler
*/
/*****************************************************************************/

#include <string.h>
#include "baseop.h"
#include "chsim.h"
#include "chsim_param.h"




/*****************************************************************************
Internal definitions
*****************************************************************************/

//! Channel simulator default configuration
static const tCHSIM_cfg CHSIM_cfgDef = {
	100,				//	100 ms	- channel delay in ms
	32767,				//	0 db	- channel attenuation
	CHSIM_CHT_FLAT,		// flat		- channel type
	0,					//  0 Hz	- currier offset in Hz
	3,					// -80 db	- 2-nd order intermodulation distortion 
	3,					// -80 db	- 3-rd order intermodulation distortion
	3,					// -80 dB	- AWGN attenuation
	3,					// -80 db	- near echo attenuation
	3,					// -80 db	- far echo attenuation
	0,					// 0 ppm	- baud rate offset in ppm
	{ 0 },				// not initialized - Near echo FIR coefficients
	{ 0 },				// not initialized - Far echo FIR coefficients
	{ 0 },				// not initialized - Channel FIR coefficients
};



/******************************************************************************
Forward and inline declarations
******************************************************************************/

//! Config and clear data
static void	CHSIM_config( tCHSIM_inst	* this );


//! Currier offset
static void CHSIM_currOffs(	tCHSIM_currOffs	* pCurr, 
							fract2q13		* pSamp, uint16		len	);

//! Intermodulation distortion calculation
static void CHSIM_intermDist(	fract2q13 * pSamp,	uint16		len,
								fract0q15	coef2,	fract0q15	coef3 );

//! Channel delay line
static void CHSIM_chanDel(	tCHSIM_chanDel	* pDel,		uint16		delMs,	
							fract2q13		* pSamp,	uint16		len		);


//! Channel distortions without near echo
/*! far echo, add WGN, currier offset, intermodulation distortion, 
	linear distortion, delay
*/
static void CHSIM_chanDist( tCHSIM_half	  * pHalf,	const tCHSIM_cfg	* pCfg,
							fract2q13	  * pSamp,	uint16				len			);

//! Near echo
static void	CHSIM_nearEcho( tCHSIM_half	  * pHalf,	const tCHSIM_cfg	* pCfg,
							fract2q13	  * pSamp,	const fract2q13		* pBack,
							uint16			len		); 


//! Resampler
static uint16 CHSIM_resamp( tRESAMP_inst	* pRes,		fract2q13 * pTo,
							const fract2q13 * pFrom,	uint16 len		);


/*****************************************************************************
External functions
*****************************************************************************/

//! Init channel simulator instance
/*! Init channel simulator instance, must be called before any use of channel simulator.
	Clear channel simulator internal data.
	\param	this			instance
	\return					true if success
*/
bool		CHSIM_init(		tCHSIM_inst * this )
{
	// init default configuration
	this->cfg = CHSIM_cfgDef;
	CHSIM_config( this );
	return true;
}


//! Start new session.
/*! Start new session: print out configuration,
	clear all internal data.
	\param	this			instance
*/
void		CHSIM_staNew(	tCHSIM_inst * this )
{
	CHSIM_printCfg( this );
	CHSIM_config( this );
}


//! Process samples.
/*! Take transmit samples of both modems, process them and
	produce receive samples for both modems.
	Number of samples for modem0 is len, number of samples 
	for modem1 is return value, it is from len-1 to len+1,
	it is unknow at call of function. Caller should always
	supply len+1 samples.
	\param	this			instance
	\param	len				number of samples for modem0
	\param	pTr0	in		modem0 output samples
	\param	pRc0	out		modem0 input/output samples
	\param	pTr1	in		modem1 output samples
	\param	pRc1	out		modem1 input/output samples
	\return					number of samples for modem1
*/
uint16	CHSIM_proc(	tCHSIM_inst		* this,	uint16		  len, 
					const fract2q13	* pTr0,	fract2q13	* pRc0,
					const fract2q13	* pTr1,	fract2q13	* pRc1	   )
{
	uint16			part, part1, part2, len1;
	fract2q13		aTmp[CHSIM_FR_SZ];

	len1 = 0;
	while( len > 0 ) {
		part = umin( len, CHSIM_FR_SZ );
	
		memcpy( pRc1, pTr0, part*sizeof(int16) );
		part1 = CHSIM_resamp( &this->aHalf[0].res, pRc0, pTr1, part );

		// from Tr0 to Rc1
		CHSIM_chanDist( &this->aHalf[0], &this->cfg, pRc1, part );
		memcpy( this->aHalf[1].aPrevRc, pRc1, part*sizeof(int16) );
		CHSIM_nearEcho( &this->aHalf[0], &this->cfg, pRc1, pRc0, part ); // Rc0 is resampled Tr1

		// from Tr1 to Rc0
		CHSIM_chanDist( &this->aHalf[1], &this->cfg, pRc0, part );
		memcpy( this->aHalf[0].aPrevRc, pRc0, part*sizeof(int16) );
		CHSIM_nearEcho( &this->aHalf[1], &this->cfg, pRc0, pTr0, part ); 

		memcpy( aTmp, pRc1, part*sizeof(int16) );
		part2 = CHSIM_resamp( &this->aHalf[1].res, pRc1, aTmp, part1 );
		ASSERT( part2 == part );

		pTr0 += part;	pRc0 += part;
		pTr1 += part1;	pRc1 += part1;
		len -= part;
		len1 += part1;
	}

	return len1;
}


/******************************************************************************
Internal functions
******************************************************************************/

//! Config and clear data
static void	CHSIM_config( tCHSIM_inst	* this )
{
	uint16			i;
	tCHSIM_half	  * pHalf;

	// clear memory
	memset( this->aHalf,	0, sizeof(this->aHalf)	  );

	// init components
	for( i = 0; i < 2; i++ ) {
		pHalf = &this->aHalf[i];
		FIR_init( &pHalf->nearFir,		CHSIM_ECHO_FIR_SZ,
				  CHSIM_aNearEchoCoef,	pHalf->aNearBuf		);
		FIR_init( &pHalf->farFir,		CHSIM_ECHO_FIR_SZ,
				  CHSIM_aFarEchoCoef,	pHalf->aFarBuf		);
		WGNGEN_init( &pHalf->wgng, 0 );
		HILB_init( &pHalf->currOf.hilb, pHalf->currOf.aBuf );
		pHalf->currOf.cph = 0;
		pHalf->currOf.dph = (ufract0q16)hir( ulfdiv( this->cfg.currOffs, (uint32)CHSIM_SAMP_RATE<<10 ) ); // currOffs in 5q10
		FIR_init( &pHalf->chanFir,		CHSIM_CHAN_FIR_SZ,
				  CHSIM_aChanCoef[0],	pHalf->aChanBuf		);
	}
	RESAMP_init( &this->aHalf[0].res,	1000000+this->cfg.baudOffs,		1000000 ); //baudOffs in ppm
	RESAMP_init( &this->aHalf[1].res,	1000000-this->cfg.baudOffs,		1000000 );

	// set fir data
	memcpy( this->cfg.aNearCoef, CHSIM_aNearEchoCoef,				CHSIM_ECHO_FIR_SZ*sizeof(int16) );
	memcpy( this->cfg.aFarCoef,  CHSIM_aFarEchoCoef,				CHSIM_ECHO_FIR_SZ*sizeof(int16) );
	memcpy( this->cfg.aChanCoef, CHSIM_aChanCoef[this->cfg.chType],	CHSIM_CHAN_FIR_SZ*sizeof(int16) );
}



//! Currier offset
static void CHSIM_currOffs(	tCHSIM_currOffs	* pCurr, 
							fract2q13		* pSamp, uint16		len  )
{
	uint16		i;
	comp16		tmp;
	comp16		aComp[CHSIM_FR_SZ];
	fract2q13	* pTo;
	const int16	invSqrt2 = 23170; // 1/sqrt(2)
	
	pTo = HILB_getPtrToWrite( &pCurr->hilb );
	memcpy( pTo, pSamp, len*sizeof(int16) );
	HILB_procBlock( &pCurr->hilb, len, aComp );
	for( i = 0; i < len; i++ ) {
		tmp = compSinCos( &pCurr->cph, pCurr->dph );
		tmp = compSmpy( aComp[i], tmp );
		tmp = compScale( tmp, invSqrt2 );
		*pSamp++ = _sadd( tmp.re, tmp.im );
	}
}



//! Intermodulation distortion calculation
void CHSIM_intermDist(	fract2q13 * pSamp,	uint16		len, 
						fract0q15	coef2,	fract0q15	coef3 )
{
	int32	sum, t32;
	int16	samp, tmp;

	// out = in + att2*in^2 att3*in^3
	for(	; len > 0; len-- ) {
		samp = *pSamp;
		sum = (int32)samp<<16;
		t32 = _lsmpy( samp, samp ); // = samp^2
		tmp = hir( _lsshl( t32, 2 ) ); //samp in 2q13
		sum = _smac( sum, tmp, coef2 );
		t32 = _lsmpy( tmp, samp ); // = samp^3
		tmp = hir( _lsshl( t32, 2 ) ); //samp in 2q13
		sum = _smac( sum, tmp, coef3 );
		*pSamp++ = hir( sum );
	}
}

//! Channel delay line
static void CHSIM_chanDel(	tCHSIM_chanDel	* pDel,		uint16	delMs, 
							fract2q13		* pSamp,	uint16	len		)
{
	uint16	i, part, ind;
	int16	* pBuf;

	ASSERT( (CHSIM_MAX_DEL_SAMP + CHSIM_FR_SZ) < MAX_UINT16 );

	// put samples
	ind = pDel->wrInd;
	part = umin( CHSIM_MAX_DEL_SAMP - ind, len );
	// from current position to end of buffer
	if( part > 0 ) {
		pBuf = &pDel->aBuf[ind];
		for( i = 0; i < part; i++ )
			*pBuf++ = *pSamp++;
	}
	part = len - part;
	// form start of buffer to end of frame
	if( part > 0 ) {
		pBuf = &pDel->aBuf[0];
		for( i = 0; i < part; i++ )
			*pBuf++ = *pSamp++;
	}
	i = ind; // save nonincremented wrInd
	ind += len;
	if( ind >= CHSIM_MAX_DEL_SAMP )
		ind -= CHSIM_MAX_DEL_SAMP;
	pDel->wrInd = ind;

	// get samples
	pSamp -= len;
	ind = i;  // nonincremented wrInd
	ind = ind + CHSIM_MAX_DEL_SAMP - ( delMs * CHSIM_SAMP_PER_MS );
	if( ind >= CHSIM_MAX_DEL_SAMP )
		ind -= CHSIM_MAX_DEL_SAMP;
	part = umin( CHSIM_MAX_DEL_SAMP - ind, len );
	// from current position to end of buffer
	if( part > 0 ) {
		pBuf = &pDel->aBuf[ind];
		for( i = 0; i < part; i++ )
			*pSamp++ = *pBuf++;
	}
	part = len - part;
	// form start of buffer to end of frame
	if( part > 0 ) {
		pBuf = &pDel->aBuf[0];
		for( i = 0; i < part; i++ )
			*pSamp++ = *pBuf++;
	}
}

//! Channel distortions without near echo
/*! far echo, add WGN, currier offset, intermodulation distortion, 
	linear distortion, delay
*/
static void CHSIM_chanDist( tCHSIM_half	  * pHalf,	const tCHSIM_cfg	* pCfg,
							fract2q13	  * pSamp,	uint16				len		)
{
	fract2q13	  * ptr;
	fract2q13		aTmp[CHSIM_FR_SZ];

	// Far echo from PrevRc to pSamp
	ptr = FIR_getPtrToWrite( &pHalf->farFir );
	memcpy( ptr, pHalf->aPrevRc, len*sizeof(int16) );
	FIR_procBlock( &pHalf->farFir, len, aTmp );
	vectScale( aTmp, pCfg->farAtt, len );
	vectSadd(  pSamp, aTmp, len );

	// Add WGN
	WGNGEN_blGen( &pHalf->wgng, len, aTmp );
	vectScale( aTmp, pCfg->awgnAtt, len );
	vectSadd(  pSamp, aTmp, len );

	// Currier offset
	CHSIM_currOffs( &pHalf->currOf, pSamp, len );

	// Intermodulation distortion
	CHSIM_intermDist( pSamp, len, pCfg->interm2, pCfg->interm3 );

	// Channel linear distortion
	ptr = FIR_getPtrToWrite( &pHalf->chanFir );
	memcpy( ptr, pSamp, len*sizeof(int16) );
	FIR_procBlock( &pHalf->chanFir, len, pSamp );
	vectScale( pSamp, pCfg->chAtt, len );

	// Channel delay
	CHSIM_chanDel( &pHalf->del, pCfg->chDel, pSamp, len );
}


//! Near echo
static void	CHSIM_nearEcho( tCHSIM_half	  * pHalf,	const tCHSIM_cfg	* pCfg,
							fract2q13	  * pSamp,	const fract2q13		* pBack,
							uint16			len		)
{
	fract2q13	  * ptr;
	fract2q13		aTmp[CHSIM_FR_SZ];

	ptr = FIR_getPtrToWrite( &pHalf->nearFir );
	memcpy( ptr, pBack, len*sizeof(int16) );
	FIR_procBlock( &pHalf->nearFir, len, aTmp );
	vectScale( aTmp, pCfg->nearAtt, len );
	vectSadd(  pSamp, aTmp, len );
}

//! Resampler
static uint16 CHSIM_resamp( tRESAMP_inst	* pRes,		fract2q13 * pTo,
							const fract2q13 * pFrom,	uint16 len		)
{
#if 0
	// not yet implemented
	uint16		i, n, len1;
	comp16		aBuf[2];

	len1  = 0;
	for(	; len > 0; len-- ) {
		aBuf[0].re = *pFrom++; 
		n = RESAMP_proc( pRes, aBuf );
		ASSERT( n <= 2 );
		for( i = 0; i < n; i++ )
			*pTo++ = aBuf[i].re;
		len1 += n;
	}
	return len1;
#else
	uint16		i;
	for( i = 0; i < len; i++ )
		*pTo++ = *pFrom++;
	return len;
#endif
}
