/****************************************************************************/
/*! \file		chsim.h
	\brief		Phone line channel simulator
	\author		Iliya Voronov, http://powerdsp.narod.ru
	\version	0.0
     
	Phone line model for modem development.
	Model simulates following channel distortions:
	1. delay and attenuation
	2. linear distortion
	3. currier offset
	4. intermodulation distortion
	5. additive wight Gaussian noise
	7. near and far echo
	8. baud rate offset

	Model process samples frame by frame, sample format is int16, 
	max input value is -8192 to 8191. Frame size is CHSIM_FR_SZ 
	at Modem0 side and CHSIM_FR_SZ +/- 1 at Modem1 side.

	Model should be initialized with CHSIM_init, 
	then model parameters should be configured by calling CHSIM_parseStr
	for each string from configuration file,
	then model can be used to process samples with CHSIM_proc.

	Channel parameters are defined in file chsim_param.c
	Line parsing commands are define in chsim_parse.c
*/
/*****************************************************************************/

#ifndef _CHSIM_H_
#define _CHSIM_H_

#include "comtypes.h"
#include "fir.h"
#include "wgngen.h"
#include "hilb.h"
#include "resamp.h"

/*****************************************************************************
Fundamental definitions
*****************************************************************************/

//! Sample rate
#define CHSIM_SAMP_RATE 8000 

//! Delay line length in ms
#define CHSIM_DEL_LEN_MS 1000

//! Frame size at modem0 side, at modem1 side it is CHSIM_FR_SZ +/- 1
#define CHSIM_FR_SZ 80


/*****************************************************************************
Internal definitions
*****************************************************************************/

//! Echo FIR filter size
#define CHSIM_ECHO_FIR_SZ 32

//! Channel linear distortion FIR in samples
#define CHSIM_CHAN_FIR_SZ 32

//! Samples per ms
#define CHSIM_SAMP_PER_MS (CHSIM_SAMP_RATE/1000)

//! Max delay line length in samples
#define CHSIM_MAX_DEL_SAMP (CHSIM_DEL_LEN_MS*CHSIM_SAMP_PER_MS)


//! Channel (linear distortion) types
typedef enum tagCHSIM_chanType {
	CHSIM_CHT_FLAT,		//!< flat
	CHSIM_CHT_GOOD,		//!< good
	CHSIM_CHT_MIDD,		//!< middle
	CHSIM_CHT_POOR,		//!< poor
	CHSIM_CHT_NUM		//!< types number
} eCHSIM_chanType;


//! Carrier offset module
typedef struct tagCHSIM_currOffs {
	tHILB_inst	hilb;	//!< hilber transform
	int16		aBuf[ HILB_TAP_SZ+CHSIM_FR_SZ-1 ];	//!< hilbert buffer
	ufract0q16	cph;	//!< current sine phase, 0 - 2pi
	ufract0q16	dph;	//!< delta phase per sample, 0 - 2pi
} tCHSIM_currOffs;

//! Channel delay buffer
typedef struct tagCHSIM_chanDel {
	int16		aBuf[ CHSIM_MAX_DEL_SAMP ];	//!< circular delay buffer
	uint16		wrInd;	//!< write index
} tCHSIM_chanDel;


//! Channel simulator data for single direction
typedef struct tagCHSIM_half {
	tFIR_inst	nearFir;		//!< near echo FIR
	int16		aNearBuf[ CHSIM_ECHO_FIR_SZ+CHSIM_FR_SZ-1 ];	//!< near echo FIR buffer
	tFIR_inst	farFir;			//!< far echo FIR
	int16		aFarBuf[ CHSIM_ECHO_FIR_SZ+CHSIM_FR_SZ-1  ];	//!< far echo FIR buffer
	tWGNGEN_inst	wgng;		//!< white gaussian noise generator
	tCHSIM_currOffs currOf;		//!< carrier offset
	tFIR_inst	chanFir;		//!< channel FIR
	int16		aChanBuf[ CHSIM_CHAN_FIR_SZ+CHSIM_FR_SZ-1 ];	//!< channel FIR buffer
	tCHSIM_chanDel	del;		//!< delay line
	int16		aPrevRc[CHSIM_FR_SZ];	//!< previous Rc frame for far echo calculation
	tRESAMP_inst	res;		//!< resampler
} tCHSIM_half;

//! Channel simulator configuration
typedef struct tagCHSIM_cfg {
	int16		chDel;		//!< channel delay in ms
	fract0q15	chAtt;		//!< channel attenuation
	eCHSIM_chanType	chType;	//!< channel type
	fract5q10	currOffs;	//!< currier offset in Hz
	fract0q15	interm2;	//!< 2-nd order intermodulation distortion 
	fract0q15	interm3;	//!< 3-rd order intermodulation distortion
	fract0q15	awgnAtt;	//!< AWGN attenuation
	fract0q15	nearAtt;	//!< near echo attenuation
	fract0q15	farAtt;		//!< far echo attenuation
	int16		baudOffs;	//!< baud rate offset in ppm
	int16		aNearCoef[ CHSIM_ECHO_FIR_SZ ];		//!< Near echo FIR coefficients
	int16		aFarCoef[  CHSIM_ECHO_FIR_SZ ];		//!< Far echo FIR coefficients
	int16		aChanCoef[ CHSIM_CHAN_FIR_SZ ];		//!< Channel FIR coefficients
} tCHSIM_cfg;


/*****************************************************************************
Object instance
*****************************************************************************/

//! Channel simulator instance
typedef struct tagCHSIM_inst {
	tCHSIM_half	aHalf[2];			//!< data for single direction
	tCHSIM_cfg	cfg;				//!< configuration
} tCHSIM_inst;



/*****************************************************************************
External functions
*****************************************************************************/

//! Init channel simulator instance
/*! Init channel simulator instance, must be called before any use of channel simulator.
	Clear channel simulator internal data.
	\param	this			instance
	\return					true if success
*/
extern bool		CHSIM_init(		tCHSIM_inst * this );

//! Start new session.
/*! Start new session: print out configuration,
	clear all internal data.
	\param	this			instance
*/

extern void		CHSIM_staNew(	tCHSIM_inst * this );



//! Parse configuration string
/*! Parse single string from configuration file to set 
	internal configuration.Argument is pure string 
	in low case, without comments etc.
	\param	this			instance
	\param	pStr	in		string to parse
	\return					true if string parsed success
*/
extern bool		CHSIM_parseStr(	tCHSIM_inst * this,
							    const char	* pStr		);

//! Print out parameters of session
/*! Print out parameters of session, to be started.
	\param	this			instance
*/
extern void		CHSIM_printCfg(	tCHSIM_inst * this );



//! Process samples.
/*! Take transmit samples of both modems, process them and
	produce receive samples for both modems.
	Number of samples for modem0 is len, number of samples 
	for modem1 is return value, it is from len-1 to len+1,
	it is unknow at call of function. Caller should always
	supply len+1 samples.
	\param	this			instance
	\param	len				number of samples for modem0
	\param	pTr0	in		modem0 output samples
	\param	pRc0	out		modem0 input/output samples
	\param	pTr1	in		modem1 output samples
	\param	pRc1	out		modem1 input/output samples
	\return					number of samples for modem1
*/
extern uint16	CHSIM_proc(	tCHSIM_inst		* this,	uint16		  len, 
							const fract2q13	* pTr0,	fract2q13	* pRc0,
							const fract2q13	* pTr1,	fract2q13	* pRc1	   );





#endif //#ifndef _CHSIM_H_



