/****************************************************************************/
/*! \file		chsim_parse.c
	\brief		Channel simulater command parser
	\author		Iliya Voronov, http://powerdsp.narod.ru
	\version	0.0
     
	Channel simulater command parcer.
*/
/*****************************************************************************/

#include <string.h>
#include <stdlib.h>
#include <math.h>

#include "baseop.h"
#include "chsim.h"
#include "chsim_param.h"
#include "fileio.h"



/*****************************************************************************
Internal definitions
*****************************************************************************/

//! Command type
typedef enum tagCHSIM_cmdType {
	CHSIM_CMDT_ENUM,		//!< text name enumerator with specified list of value 
	CHSIM_CMDT_FRCT,		//!< float value to convert to fract16 with specified divider
	CHSIM_CMDT_DB			//!< float value in dB for conversion to fractiona number
} eCHSIM_cmdType;


//! Command processing functions definition
typedef bool (* fCHSIM_cmdProc) ( const void * pDesc, const char * pArg, tCHSIM_cfg * pCfg );


//! Command description
typedef struct tagCHSIM_cmdDesc {
//	eCHSIM_cmdType		type;	//!< command type
	const char		  *	pName;	//!< command name
	fCHSIM_cmdProc		fProc;	//!< processing function
	double				min;	//!< min argument raw value
	double				max;	//!< max argument raw value
	int16 			  *	pVal;	//!< offset of field of tCHSIM_cfg to write value
	int16				mult;	//!< multiplier to convert double to fractional, for CHSIM_CMDT_FRCT only
} tCHSIM_cmdDesc;




/******************************************************************************
Forward and inline declarations
******************************************************************************/

//! Enumerator command processing function  
static bool	CHSIM_cmdChtype( const tCHSIM_cmdDesc * pDesc, const char * pArg, tCHSIM_cfg * pCfg );

//! Fractional command processing function  
static bool	CHSIM_cmdFrct(	 const tCHSIM_cmdDesc * pDesc, const char * pArg, tCHSIM_cfg * pCfg );

//! dB command processing function  
static bool	CHSIM_cmdDb(	 const tCHSIM_cmdDesc * pDesc, const char * pArg, tCHSIM_cfg * pCfg );



/******************************************************************************
Command description
******************************************************************************/

//! Command description
static const tCHSIM_cmdDesc CHSIM_aCmdDesc[] = {
//		pName		fProc			 min	  max		pVal								mult
	{	"chdel",	CHSIM_cmdFrct,	  0.,	2000.,		&(((tCHSIM_cfg*)NULL)->chDel),		   1	},
	{	"chatt",	CHSIM_cmdDb,	  0.,	  60.,		&(((tCHSIM_cfg*)NULL)->chAtt),				},
	{	"chtype",	CHSIM_cmdChtype																	},
	{	"currof",	CHSIM_cmdFrct,	-30.,	 +30.,		&(((tCHSIM_cfg*)NULL)->currOffs),	1024	},
	{	"interm2",	CHSIM_cmdDb,	 10.,	  80.,		&(((tCHSIM_cfg*)NULL)->interm2),			},
	{	"interm3",	CHSIM_cmdDb,	 10.,	  80.,		&(((tCHSIM_cfg*)NULL)->interm3),			},
	{	"snr",		CHSIM_cmdDb,	  0.,	  80.,		&(((tCHSIM_cfg*)NULL)->awgnAtt),			},
	{	"nearatt",	CHSIM_cmdDb,	  6.,	  80.,		&(((tCHSIM_cfg*)NULL)->nearAtt),			},
	{	"faratt",	CHSIM_cmdDb,	  6.,	  80.,		&(((tCHSIM_cfg*)NULL)->farAtt),				},
//	{	"baudof",	CHSIM_cmdFrct, -200.,	+200.,		&(((tCHSIM_cfg*)NULL)->baudOffs),	   1	},
	{	NULL																						}
};


//! Argument description - chtype - enum name 
static const char * const CHSIM_apArgChtype[ CHSIM_CHT_NUM ] = {
	"flat", 
	"good", 
	"midd", 
	"poor"
};


/*****************************************************************************
External functions
*****************************************************************************/

//! Parse configuration string
/*! Parse single string from configuration file to set 
	internal configuration. Argument is pure string 
	in low case, without comments etc.
	\param	this			instance
	\param	pStr	in		string to parse
	\return					true if string parsed success
*/
bool	CHSIM_parseStr(	tCHSIM_inst * this,
						const char	* pStr		)
{
	const tCHSIM_cmdDesc	* pCmd;
	const char				* pArg;
	size_t					len;

	pStr += strspn( pStr, " \t\n" );	// find first token - command name

	for ( pCmd = CHSIM_aCmdDesc; pCmd->pName != NULL; pCmd++ ) {
		len = strlen( pCmd->pName );
		if ( strncmp( pStr, pCmd->pName, len ) == 0 ) { // Token found in list of command
			pArg = pStr + len;
			pArg += strspn( pArg, " \t\n" ); // find nex token - command argument
			if ( pCmd->fProc( pCmd, pArg, &this->cfg ) ) // Arguments are succesfully processed
				return true;
		}
	}
	return false;
}

//! Print out parameters of session
/*! Print out parameters of session, to be started.
	\param	this			instance
*/
void	CHSIM_printCfg(	tCHSIM_inst * this )
{
	FILEIO_printf( "  SNR %.2f dB, channel type '%s'\n", 
			20.*log10( 32768. / (double)this->cfg.awgnAtt ),
			CHSIM_apArgChtype[ this->cfg.chType ]							);
	FILEIO_printf( "  Channel attenuation: %.2f, delay: %ld ms dB\n", 
			20.*log10( 32768. / (double)this->cfg.chAtt ),
			(long)this->cfg.chDel											);
	FILEIO_printf( "  Currier offset: %.3f Hz, baud offset: %.2f ppm\n",
			(double)this->cfg.currOffs / 1024., (double)this->cfg.baudOffs	);
	FILEIO_printf( "  Echo attenuation near: %.2f dB, far: %.2f dB\n",
			20.*log10( 32768. / (double)this->cfg.nearAtt ),
			20.*log10( 32768. / (double)this->cfg.farAtt )					);
	FILEIO_printf( "  Intermodulation distortions 2-nd: %.2f dB, 3-rd: %.2f dB\n",
			20.*log10( 32768. / (double)this->cfg.interm2 ),
			20.*log10( 32768. / (double)this->cfg.interm3 )					);
}


/******************************************************************************
Internal functions
******************************************************************************/

//! Enumerator command processing function  
static bool	CHSIM_cmdChtype( const tCHSIM_cmdDesc * pDesc, const char * pArg, tCHSIM_cfg * pCfg )
{
	uint16			i;
	const char		* pName;
	size_t			len;

	for ( i = 0; i < CHSIM_CHT_NUM; i++ ) {
		pName = CHSIM_apArgChtype[i];
		len = strlen( pName );
		if ( strncmp( pArg, pName, len ) == 0 ) { // Token found in list of arguments
			pArg = pArg + len;
			pArg += strspn( pArg, " \t\n" ); // span white characters
			if( *pArg == '\0' ) { // string is fully parsed
				pCfg->chType = (eCHSIM_chanType)i;
				memcpy( pCfg->aChanCoef, CHSIM_aChanCoef[i], sizeof(pCfg->aChanCoef) );
				return true;
			}
		}
	}
	return false;
}

//! Fractional command processing function  
static bool	CHSIM_cmdFrct(	 const tCHSIM_cmdDesc * pDesc, const char * pArg, tCHSIM_cfg * pCfg )
{
	double		val;
	char		* pEnd;

	val = strtod( pArg, &pEnd ); 
	pEnd += strspn( pEnd, " \t\n" ); // span white characters
	if( *pEnd == '\0' ) // string is fully parsed
		if( ( val <= pDesc->max ) && ( val >= pDesc->min ) ) {
			val *= (double)pDesc->mult;
			*(int16*)((char*)pCfg+(size_t)pDesc->pVal) = (int16)val;
			return true;
		}
	return false;
}

//! dB command processing function  
static bool	CHSIM_cmdDb(	 const tCHSIM_cmdDesc * pDesc, const char * pArg, tCHSIM_cfg * pCfg )
{
	double		val;
	char		* pEnd;

	val = strtod( pArg, &pEnd ); 
	pEnd += strspn( pEnd, " \t\n" ); // span white characters
	if( *pEnd == '\0' ) // string is fully parsed
		if( ( val <= pDesc->max ) && ( val >= pDesc->min ) ) {
			val = pow( 10., val/-20. );
			*(int16*)((char*)pCfg+(size_t)pDesc->pVal) = sat((int32)( val * (double)0x8000UL ));
			return true;
		}
	return false;
}
