/****************************************************************************/
/*! \file		wgnGen.c
	\brief		White Gaussian noise generator.
	\author		Iliya Voronov, http://powerdsp.narod.ru
	\version	1.2
     
	White Gaussian noise generator.
	1. White Gaussian noise is generated in following way
		Linear congruent generator (m = 2^32, a = 1103515245, c = 12345) 
		is used to generate uniform number. Gaussian number is calculated
		as sum of WGNGEN_UNIF_NUM uniform number.
	2. Band limited white Gaussian noise is also generated by 
		filtering white noise with band pass filter 600 - 3000 Hz (Fs = 8 kHz)
		0.1 roll-off and 4.5 dB attenuation at the edges.
*/
/*****************************************************************************/

#include <string.h>
#include "wgnGen.h"
#include "baseop.h"


/******************************************************************************
Internal definitions
******************************************************************************/

//! Number of uniform value per gaussian
#define WGNGEN_UNIF_NUM 12

//! Linear congruent generator A
#define WGNGEN_LCG_A 1103515245UL

//! Linear congruent generator C
#define WGNGEN_LCG_C 12345UL


//! Band-limiting FIR filte coefficients
/*! It is multiband FIR filter, designed in QED2000,
	files filt/noise.spc and filt/noise.flt.
	Sample rate 8000Hz, pass band 600 - 3000 Hz
	0.1 roll-off and 4.5 dB attenuation at the edges
	Bands:	   0- 300,	0.000 +- 0.001
			   600,		0.596
			1000-1200,  1.000 +- 0.010
			  3000,		0.596
			3300-4000,  0.000 +- 0.001
	All coefficients are multiplied by 1.3010 to compensate 
	suppressed bands.
	Matlab code:
	sp = fft( coef/32768 );
	psp = abs( sp );
	psp = psp.*psp;
	avpsp = sum( psp ) / size( psp, 2 );
	mult = sqrt( 1/avpsp );
*/
static const int16 WGNGEN_aBlFirCoef[ WGNGEN_BL_FIR_SZ ] = {
    111,    164,   -211,   -211,     43,    -95,     20,    199,
      4,    206,    180,   -164,    109,   -267,   -497,     14,
   -639,    -74,    483,   -254,   1370,    812,    343,   2066,
   -972,    -51,   -285,  -5334,   -440,  -6466,  -9335,  19235,
  19235,  -9335,  -6466,   -440,  -5334,   -285,    -51,   -972,
   2066,    343,    812,   1370,   -254,    483,    -74,   -639,
     14,   -497,   -267,    109,   -164,    180,    206,      4,
    199,     20,    -95,     43,   -211,   -211,    164,    111
/* nonmultiplied
	 85,    126,   -162,   -162,     33,    -73,     15,    153,
      3,    158,    138,   -126,     84,   -205,   -382,     11,
   -491,    -57,    371,   -195,   1053,    624,    264,   1588,
   -747,    -39,   -219,  -4100,   -338,  -4970,  -7175,  14785,
  14785,  -7175,  -4970,   -338,  -4100,   -219,    -39,   -747,
   1588,    264,    624,   1053,   -195,    371,    -57,   -491,
     11,   -382,   -205,     84,   -126,    138,    158,      3,
    153,     15,    -73,     33,   -162,   -162,    126,     85 */
};


/******************************************************************************
Forward and inline declarations
******************************************************************************/

//! Calculate next LCG number 
_inline_ uint32	WGNGEN_lcgCalc( uint32 a )
{
	return a * WGNGEN_LCG_A + WGNGEN_LCG_C;
}


/******************************************************************************
External functions
******************************************************************************/

//! Initialize WGN generator.
/*! Initialize WGN generator and set seed.
  \param	this			WGN generator instance
  \param	seed			seed value	
*/
void WGNGEN_init( tWGNGEN_inst	* this,		uint32	seed  )
{
	this->lcg = seed;
	this->lcg = WGNGEN_lcgCalc( this->lcg );
	ASSERT( WGNGEN_UNIF_NUM == 12 ); // for simplicity, variance = 1/12
	memset( this->aBuf, 0, sizeof( this->aBuf ) );
	FIR_init( &this->fir, WGNGEN_BL_FIR_SZ, WGNGEN_aBlFirCoef, this->aBuf );
}


//! Generate WGN samples.
/*! Generate WGN samples with 0. mean and 1. variance
  \param	this			WGN generator instance
  \param	len				number of samples
  \param	pWgn	out		WGN samples 
*/
void	WGNGEN_gen(	tWGNGEN_inst	* this,		uint16	len, 
					fract7q23		* pWgn   )
{
	uint16		i;
	int40		acc;

	for(	; len > 0; len-- ) {
		acc = 0;
		for( i = 0; i < WGNGEN_UNIF_NUM; i++ ) {
			this->lcg = WGNGEN_lcgCalc( this->lcg );
			acc = _llsadd( acc, (int32)( this->lcg << 1 ) ); // mean = 0, var = 1/12
		}
		*pWgn++ = lsmpyls( (int32)( acc>>8 ), 23170 ); // * sqrt( WGNGEN_UNIF_NUM / (12*2) )
	}
}

//! Generate band-limited WGN samples.
/*! Generate 16-bit band-limited WGN samples with mean = 0 and variance = 1
	in band 600 - 3000 Hz (sample rate 8 kHz)
	0.1 roll-off and 4.5 dB attenuation at the edges.
  \param	this			WGN generator instance
  \param	len				number of samples
  \param	pWgn	out		WGN samples
*/
void	WGNGEN_blGen(	tWGNGEN_inst	* this,		uint16	len, 
						fract2q13		* pWgn   )
{
	uint16		  part, i;
	fract7q23	  aWgn[WGNGEN_BL_FR_SZ];
	fract2q13	* pWr;

	while( len > 0 ) {
		part = umin( len, WGNGEN_BL_FR_SZ );
		WGNGEN_gen( this, part, aWgn );
		pWr = FIR_getPtrToWrite( &this->fir );
		for( i = 0; i < part; i++ )
			*pWr++ = hir( _lsshl( aWgn[i], 5 ) );
		FIR_procBlock( &this->fir, part, pWgn );
		pWgn += part;
		len -= part;
	}
}


