/****************************************************************************/
/*! \file		wgnGen.h
	\brief		White Gaussian noise generator.
	\author		Iliya Voronov, http://powerdsp.narod.ru
	\version	1.2
     
	White Gaussian noise generator.
	1. White Gaussian noise is generated in following way
		Linear congruent generator (m = 2^32, a = 1103515245, c = 12345) 
		is used to generate uniform number. Gaussian number is calculated
		as sum of WGNGEN_UNIF_NUM uniform number.
	2. Band limited white Gaussian noise is also generated by 
		filtering white noise with band pass filter 600 - 3000 Hz (Fs = 8 kHz)
		0.1 roll-off and 4.5 dB attenuation at the edges.

*/
/*****************************************************************************/
#ifndef _WGN_GEN_H_
#define _WGN_GEN_H_

#include "comtypes.h"
#include "fir.h"

/******************************************************************************
Internal definitions
******************************************************************************/

//! Band limiting FIR filter size
#define WGNGEN_BL_FIR_SZ 64

//! Band-limited WGN frame size
#define WGNGEN_BL_FR_SZ 32


/******************************************************************************
Object instance
******************************************************************************/

//! Object instance
typedef struct tagWGNGEN_inst {
	uint32			lcg;	//!< linear congruent generator value
	tFIR_inst		fir;		//!< band-limiting FIR
	int16			aBuf[WGNGEN_BL_FIR_SZ+WGNGEN_BL_FR_SZ-1];	//!< band-limiting FIR delay line
} tWGNGEN_inst;


/******************************************************************************
External functions
******************************************************************************/

//! Initialize WGN generator.
/*! Initialize WGN generator and set seed.
  \param	this			WGN generator instance
  \param	seed			seed value	
*/
_extern_c_ void WGNGEN_init( tWGNGEN_inst	* this,		uint32	seed  );


//! Generate WGN samples.
/*! Generate 32-bit WGN samples with mean = 0 and variance = 1
	in band from 0 to Fs/2.
  \param	this			WGN generator instance
  \param	len				number of samples
  \param	pWgn	out		WGN samples 
*/
_extern_c_ void	WGNGEN_gen(	tWGNGEN_inst	* this,		uint16	len, 
							fract7q23		* pWgn   );


//! Generate band-limited WGN samples.
/*! Generate 16-bit band-limited WGN samples with mean = 0 and variance = 1
	in band 600 - 3000 Hz (sample rate 8 kHz)
	0.1 roll-off and 4.5 dB attenuation at the edges.
  \param	this			WGN generator instance
  \param	len				number of samples
  \param	pWgn	out		WGN samples
*/
_extern_c_ void	WGNGEN_blGen(	tWGNGEN_inst	* this,		uint16	len, 
								fract2q13		* pWgn   );



#endif //_WGN_GEN_H_

