/****************************************************************************/
/*! \file		baseop.h
	\brief		Base operations
	\author		Iliya D. Voronov, http://powerdsp.narod.ru
	\version	1.2
     
	Base operations. Common operations for all platforms.
*/
/*****************************************************************************/

#include "baseop.h"



/******************************************************************************
Devision, inversion
******************************************************************************/

/*! Fractional division
    deviation from floating point version +- 2 LSB
    if num >= den ( and den == 0 ), result saturated to maximal value
  \param    num   numerator in any 32-bit unsigned fractional format
  \param    den   denominator in the same unsigned fractional format as numerator
  \return         devision result, in Q32 format
*/
ufract0q32 ulfdiv(  ufract0q32 num,  ufract0q32 den )
{
  int16 i;
  ufract0q32 res;
  if ( num >= den )
    return MAX_UINT32;
  res = 0;
  { // first iteration of for( i=0; i<31; i++ )
    if ( num >= den ) {
      num -= den;
      res += 1;
    }
    ( den & 0x80000000UL ) ? ( den >>= 1 ) : ( num <<= 1 ); 
    res <<= 1;
  }
  for( i=0; i<31; i++ ) {
    if ( num >= den ) {
      num -= den;
      res += 1;
    }
    num <<= 1;
    res <<= 1;
  }
  return res;
}

#if 0
/*! Fractional division
    deviation from floating point version +- 1 LSB
    if num >= den ( and den == 0 ), result saturated to maximal value
  \param    shNum   numerator in any 16-bit unsigned fractional format
  \param    shDen   denominator in the same unsigned fractional format as numerator
  \return           devision result, in Q16 format
*/
ufract0q16 ufdiv(  ufract0q16 shNum,  ufract0q16 shDen )
{
  int16 i;
  ufract0q16 res;
  ufract0q32 num, den;
  if ( shNum >= shDen )
    return MAX_UINT16;

  num = shNum;
  den = shDen;
  res = 0;
  for( i=0; i<16; i++ )
  {
    if ( num >= den )
    {
      num -= den;
      res += 1;
    }
    num <<= 1;
    res <<= 1;
  }
  return res;
}

/*! Fractional division
    deviation from floating point version +- 1 LSB
    if abs(num) >= abs(den) ( or den == 0 ), result saturated to maximal value
  \param    shNum   numerator in any 16-bit signed fractional format
  \param    shDen   denominator in the same signed fractional format as numerator
  \return           devision result, in SQ16 format
*/
fract0q16 fdiv(  fract0q16 shNum,  fract0q16 shDen )
{
  tFract16   res;
  int16       sign = (shNum ^ shDen) & 0x8000;
  res = (tFract16)(div_us( _abss(shNum), _abss(shDen) )>>1);
  if( sign )
    res = -res;
  return res;
}
#endif


/******************************************************************************
Square root
******************************************************************************/

#if 0
/*! Square root calculation
    deviation from floating point version +- 1 LSB
  \param    arg   argument, in Q32, 2Q30, 4Q28 ... 32Q unsigned fractional format
  \return         square root value, in Q16, 1Q15, 2Q14 ... 16Q unsigned fractional format
*/ 
#if 1
ufract0q16 fsqrt( ufract0q32 arg )
{

  ufract0q16	res, i, step;
  static ufract0q16 aNew[1]; // to use pointer
  ufract0q16	* pNew = aNew; // pointer is used to allow 1 cycle mpy
  ufract0q32	sq;

  res  = 0;
  step = 0x8000;
  *pNew = res + step;
  for( i = 0; i < 16; i++ ) {
    sq  = ((uint32)(*pNew)) * ((uint32)(*pNew));
    if ( sq <= arg )
      res += step;
    *pNew = res + (step>>1);
    step >>= 1;
  }

  return res;
}
#else
ufract0q16 sqrt_ul( ufract0q32 arg )
{
  ufract0q16 res, tmp, step;
  ufract0q32 sqr;

  res  = 0;
  for( step = 0x8000; step != 0; step >>= 1 )
  {
    tmp = res + step;
    sqr = mult_usus( tmp, tmp );
    if ( sqr <= arg )
      res = tmp;
  }
  return res;
}
#endif


/*! Cubic root calculation
  \param    arg   argument, in Q32, 3Q29, 6Q26 unsigned fractional format
  \return         square root value, in Q16, 1Q15, 2Q14 unsigned fractional format
*/ 
ufract0q16 fcubrt( ufract0q32 arg )
{
  ufract0q16	res, tmp, step;
  ufract0q32	sqr;

  res  = 0;
  for( step = 0x8000; step != 0; step >>= 1 )
  {
    tmp = res + step;
    sqr = mult_ulus( mult_usus( tmp, tmp ), tmp );
    if ( sqr <= arg )
      res = tmp;
  }
  return res;
}
#endif


/******************************************************************************
Math functions
******************************************************************************/

#if 0
//! 0 to pi/4 in 
static const fractRad MATHOP_aAtanTbl[] = 
{
    0,  651, 1298, 1934, 2556, 3161, 3744, 4303,
 4838, 5346, 5828, 6284, 6714, 7118, 7499, 7857,
 8192
};


/*! Arctangent of x/y calculation
    if x and y are 0, atan2 returns 0
    deviation from floating point version +- 4 LSB
  \param    x   argument, in signed fractional format
  \param    y   argument, in the same fractional format as x
  \return       angle, 8000 to 7FFF -> -pi to pi
*/
fractRad atan2_ss ( fract0q15 x, fract0q15 y )
{
  uint16    quadr;  // quadrant: 0 - 0 to pi/2, 1 - pi/2 to pi, 2 - pi to 3*pi/2, 3 - 3*pi/2 to 2*pi
  bool		isLow;  // is low half of quadrant 0 to pi/4 (after transfer to 0 quadrant)  
  ufract0q16	ratio;  // x/y transfered to first half of first quadrant
  uint16    num;    // approximation interval number
  fractRad	res;

  if ( ( x == 0 ) && ( y == 0 ) ) // standard atan2 behaviour
    return 0;

  quadr = ( x >= 0 ) ? ( ( y >= 0 ) ? 0 : 1 ) : ( ( y >= 0 ) ? 3 : 2 );
  x = _abss( x ); // transfer to first quadrant  
  y = _abss( y );

  isLow = ( x < y );
  ratio = ( isLow ) ? ufdiv( x, y ) : ufdiv( y, x );  

  num = (ratio>>12) & 0x000F;
  ratio &= 0x0FFF;
  res  = umpyuu( MATHOP_aAtanTbl[num]   ,
				(ufract0q16)((0x0FFF-ratio)<<4) )   ; 
  res += umpyuu( MATHOP_aAtanTbl[num+1] ,
				slsh_us((ufract0q16)(ratio+1), 4)  );

  if ( ! isLow )
    res = 0x4000 - res;
  switch ( quadr )
  {
    case 0: res =          res; break;
    case 1: res = 0x8000 - res; break;
    case 2: res = 0x8000 + res; break;
    case 3: res =        - res; break;
  }
  return res;
}
#endif

// Sine table 0 to 2pi
const fract0q15 BASEOP_aSinTbl[256] = {
     0,   804,  1608,  2411,  3212,  4011,  4808,  5602,
  6393,  7180,  7962,  8740,  9512, 10279, 11039, 11793,
 12540, 13279, 14010, 14733, 15447, 16151, 16846, 17531,
 18205, 18868, 19520, 20160, 20788, 21403, 22006, 22595,
 23170, 23732, 24279, 24812, 25330, 25833, 26320, 26791,
 27246, 27684, 28106, 28511, 28899, 29269, 29622, 29957,
 30274, 30572, 30853, 31114, 31357, 31581, 31786, 31972,
 32138, 32286, 32413, 32522, 32610, 32679, 32729, 32758,
 32767, 32758, 32729, 32679, 32610, 32522, 32413, 32286,
 32138, 31972, 31786, 31581, 31357, 31114, 30853, 30572,
 30274, 29957, 29622, 29269, 28899, 28511, 28106, 27684,
 27246, 26791, 26320, 25833, 25330, 24812, 24279, 23732,
 23170, 22595, 22006, 21403, 20788, 20160, 19520, 18868,
 18205, 17531, 16846, 16151, 15447, 14733, 14010, 13279,
 12540, 11793, 11039, 10279,  9512,  8740,  7962,  7180,
  6393,  5602,  4808,  4011,  3212,  2411,  1608,   804,
     0,  -804, -1608, -2411, -3212, -4011, -4808, -5602,
 -6393, -7180, -7962, -8740, -9512,-10279,-11039,-11793,
-12540,-13279,-14010,-14733,-15447,-16151,-16846,-17531,
-18205,-18868,-19520,-20160,-20788,-21403,-22006,-22595,
-23170,-23732,-24279,-24812,-25330,-25833,-26320,-26791,
-27246,-27684,-28106,-28511,-28899,-29269,-29622,-29957,
-30274,-30572,-30853,-31114,-31357,-31581,-31786,-31972,
-32138,-32286,-32413,-32522,-32610,-32679,-32729,-32758,
-32768,-32758,-32729,-32679,-32610,-32522,-32413,-32286,
-32138,-31972,-31786,-31581,-31357,-31114,-30853,-30572,
-30274,-29957,-29622,-29269,-28899,-28511,-28106,-27684,
-27246,-26791,-26320,-25833,-25330,-24812,-24279,-23732,
-23170,-22595,-22006,-21403,-20788,-20160,-19520,-18868,
-18205,-17531,-16846,-16151,-15447,-14733,-14010,-13279,
-12540,-11793,-11039,-10279, -9512, -8740, -7962, -7180,
 -6393, -5602, -4808, -4011, -3212, -2411, -1608,  -804,
};

#if 0
/*! Cosine calculation
    deviation from floating point version +- 4 LSB
  \param  arg   angle, 8000 to 7FFF -> -pi to pi
  \return       cosine value in SQ15 format       
*/
fract0q15 cos_s ( fractRad arg )
{
  uint16    quadr;  // quadrant: 0 - 0 to pi/2, 1 - pi/2 to pi, 2 - pi to 3*pi/2, 3 - 3*pi/2 to 2*pi
  uint16    num;    // approximation interval number
  fract0q15	res;

  quadr = ((uint16)arg)>>14;

  switch ( quadr ) 
  {
    case 0: arg =          arg; break; // transfer to first quadrant
    case 1: arg = 0x7FFF - arg; break;
    case 2: arg = 0x8000 + arg; break;
    case 3: arg = 0xFFFF - arg; break;
  }

  arg  = 0x3FFF - arg; // calculate cosine by sine tabel

  num  =  (arg>>8) & 0x003F;
  arg &= 0x00FF;

  res  = umpyuu( MATHOP_aSinTbl[num]   ,
				(ufract0q16)((0x00FF-arg)<<8) ); 
  res += umpyuu( MATHOP_aSinTbl[num+1] ,
				slsh_us((ufract0q16)(arg+1), 8) );

  switch ( quadr )
  {
    case 0:
    case 3:   res =   res; break;
    case 1:
    case 2:   res = - res; break;
  }
  return res;
}
#endif



/******************************************************************************
Vector operations
******************************************************************************/

void vectSadd( int16 * pR, const int16 * pA, uint16 len )
{
	while( len-- > 0 ) {
		*pR = _sadd( *pR, *pA );
		pR++;
		pA++;
	}
}


void vectSmpy( int16 * pR, const int16 * pA, uint16 len )
{
	while( len-- > 0 ) {
		*pR = _smpy( *pR, *pA );
		pR++;
		pA++;
	}
}


void vectScale( int16 * pR, int16 a, uint16 len )
{
	while( len-- > 0 ) {
		*pR = _smpy( *pR, a );
		pR++;
	}
}
