/****************************************************************************/
/*! \file		basepop.h
	\brief		Base operations
	\author		Iliya Voronov, http://powerdsp.narod.ru
	\version	1.0
     
	Base operations. Common operations for all platforms.
*/
/*****************************************************************************/

#ifndef _BASEOP_H_
#define _BASEOP_H_

#include "comtypes.h"
#include "intrins5X.h"


/******************************************************************************
General functions
******************************************************************************/
_inline_ uint16 lo(  int32 a )		{	return (uint16)a;					}
_inline_ int16  hi(  int32 a )		{	return (int16)( a >> 16 );			}
_inline_ int16  hir( int32 a )		{	return (int16)( _rnd(a) >> 16 );	}

_inline_ int16   sshr( int16 a, int16 s )		{	return   _shrs( a, s );		}

_inline_ uint16  umax( uint16 a, uint16 b )		{	return   (a > b) ? a : b;	}
_inline_ uint32 ulmax( uint32 a, uint32 b )		{	return   (a > b) ? a : b;	}
_inline_ uint16  umin( uint16 a, uint16 b )		{	return   (a < b) ? a : b;	}
_inline_ uint32 ulmin( uint32 a, uint32 b )		{	return   (a < b) ? a : b;	}

_inline_ int16   sat( int32 a )		{	return (int16)_lmin( _lmax( a, (int32)MIN_INT16 ), (int32)MAX_INT16 );	}

_inline_ int16	bcnt( uint40 a )	{	return  _count( a, a );		}

_inline_ uint32 ulswap( uint32 a )	{	return( a << 16 ) | ( a >> 16 );	}


/******************************************************************************
Simple operations
******************************************************************************/

_inline_ int32 lsmpyls( int32 a, int16 b )
{
  int16  h;
  uint16 l;
  int32  Y1, Y2;

  h =  (int16)((a >>16) & 0xFFFF);
  l =  (uint16)(a & 0xFFFF);

  Y1 =  ((int32)h * b) << 1;
  Y2 =  ((int32)l * b + 0x4000) >> 15;
  return Y1 + Y2;
}


_inline_ int32 lsmpyll ( int32 L_var1, int32 L_var2 )
{
  int16 swLow1,
        swLow2,
        swHigh1,
        swHigh2;
  int32 L_mid;

  swLow1 = (((int16)(L_var1))>>1)&0x7FFF;
  swLow2 = (((int16)(L_var2))>>1)&0x7FFF;
  swHigh1 = hi(L_var1);
  swHigh2 = hi(L_var2);

  L_mid = (_lsmpy(swHigh1, swLow2)>>1) +
          (_lsmpy(swHigh2, swLow1)>>1) +
          (_lsmpy(swLow1,  swLow2)>>16);
  L_mid = L_mid >> 14;
  L_mid += _lsmpy(swHigh1, swHigh2);

  return L_mid;
}

#if 0
//! Multiply 32x32, return 32MSB
__inline int32 mpyll( int32 a, int32 b )
{
	int32 r;
	r = _mpyh( a, b );
	r = _sadd( r, _mpyhslu( a, b )>>16 );
	r = _sadd( r, _mpyluhs( a, b )>>16 );
	return _sshl( r, 1 );
}
#endif



//! Unsigned long division for fractional format 
_extern_c_ ufract0q32 ulfdiv( ufract0q32 a, ufract0q32 b );

#if 0
//! Unsigned division for fractional format 
_extern_c_ ufract0q16 ufdiv( ufract0q16 a, ufract0q16 b );

//! Signed division for fractional format 
//_extern_c_ fract0q15 fdiv( fract0q15 a, fract0q15 b );
#endif

/******************************************************************************
Math functions
******************************************************************************/
#if 0
//! Square root for fractional format 
_extern_c_ ufract0q16 fsqrt( ufract0q32 arg );

//! Cubic root for fractional format 
//_extern_c_ ufract0q16 fcubrt( ufract0q32 arg );

//! x / y arctangent
_extern_c_ fractRad atan2_ss ( fract0q15 x, fract0q15 y );

//! Cosine
_extern_c_ fract0q15 cos_s ( fractRad arg );
#endif

/******************************************************************************
Complex operations
******************************************************************************/

_inline_ comp16 compSadd( comp16 a, comp16 b)
{
  comp16 res;
  res.re = _sadd ( a.re, b.re );
  res.im = _sadd ( a.im, b.im );
  return res;
}

_inline_ comp16 compSsub( comp16 a, comp16 b)
{
  comp16 res;
  res.re = _ssub ( a.re, b.re );
  res.im = _ssub ( a.im, b.im );
  return res;
}


_inline_ comp16 compSshr( comp16 a, int16 s)
{
  comp16 res;
  res.re = _shrs ( a.re, s );
  res.im = _shrs ( a.im, s );
  return res;
}

_inline_ comp16 compSmpy( comp16 a, comp16 b)
{
  comp16 res;
  int32   re, im;
  re = _lsmpy( a.re, b.re );
  re = _smas ( re, a.im, b.im );
  im = _lsmpy( a.re, b.im );
  im = _smac ( im, a.im, b.re );
  res.re = hi ( re );
  res.im = hi ( im );
  return res;
}

_inline_ comp16 compScale( comp16 a, int16 b )
{
  comp16 res;
  res.re = _smpy( a.re, b );
  res.im = _smpy( a.im, b );
  return res;
}

_inline_ int32 compSqr( comp16 a )
{
  int32 res;
  res = _lsmpy( a.re, a.re );
  res = _smac ( res, a.im, a.im );
  return res;
}


extern const fract0q15 BASEOP_aSinTbl[];

/*! Sin/Cosin fast calculation
    deviation from floating point version +- 4 LSB
  \param	pCph	in/out	pointer to cuttent phase, 0 to 2pi
  \param	pph				delat phase, 0 to 2pi
  \return					cosine value in SQ15 format       
*/
_inline_ comp16 compSinCos( ufract0q16 * pCph, ufract0q16 dph )
{
	uint16	ind;
	comp16	res;

	ind = (*pCph)>>8;
	*pCph += dph;
	res.im = BASEOP_aSinTbl[ind];
	ind = (ind+64)&0x00FF;
	res.re = BASEOP_aSinTbl[ind];
	return res;
}


/******************************************************************************
Vector operations
******************************************************************************/

_extern_c_ void vectSadd( int16 * pR, const int16 * pA, uint16 len );

_extern_c_ void vectSmpy( int16 * pR, const int16 * pA, uint16 len );

_extern_c_ void vectScale( int16 * pR, int16 a, uint16 len );

#endif //#ifndef _BASEOP_H_



