/****************************************************************************/
/*! \file		fir.c
	\brief		FIR filter
	\author		Iliya Voronov, http://powerdsp.narod.ru
	\version	1.2
     
	FIR filter, 16bit precision.
*/
/*****************************************************************************/


#include <string.h>
#include "fir.h"

#include "baseop.h"


/******************************************************************************
External functions
******************************************************************************/

//! Initialize FIR filter object.
/*! Clear buffer and set all parameters. It should be called 
    befor any other functions.
  \param  this          pointer to FIR filter instance
  \param  filtSize      filter size (number of tap)
  \param  pCoef    in   pointer to filter coefficients 
  \param  pBuf     in   pointer to buffer  
*/
void FIR_init(  tFIR_inst  * this, 
                int16           filtSize, 
                const int16   * pCoef,
                int16         * pBuf      )
{
	this->filtSize    = filtSize;

	ASSERT( pCoef );
	this->pCoef       = pCoef;
	ASSERT( pBuf );
	this->pBuf       = pBuf;
}




//! Process FIR filter calculation for one block.
/*! This function run FIR with the internal data and write output 
    blockSize sampels by the pOut pointer. Function processes 
    blockSize input samples, which should be previousely written 
    by pointer, returned by FIR_getPtrToWrite. 
  \param    this		pointer to FIR filter  instance
  \param	blockSize	number of samples, processed per FIR_procBlock call 
  \param    pOut		out pointer to write to output samples
*/
void FIR_procBlock( tFIR_inst * this,	int16	blockSize,
                    int16     * pOut   )
{
	uint16     	i, j;
	const int16	* pIn;
	const int16	* pCoef;
	int32			sum;

#ifdef COMPILER_C55
	#pragma MUST_ITERATE(1);
#endif //#ifdef COMPILER_C55
	for ( i=0; i < blockSize; i++ ) { // Output block process loop 
		pIn   = &this->pBuf[i];
		pCoef = this->pCoef;
		sum = 0;

#ifdef COMPILER_C55
		#pragma MUST_ITERATE(1);
#endif //#ifdef COMPILER_C55
		for ( j=0; j < this->filtSize; j++ ) // FIR loop
			sum = _smac( sum, *pIn++, *pCoef++);
		*pOut++ = hir( sum );
	}
	memcpy(	&this->pBuf[0],
			&this->pBuf[blockSize],
			(this->filtSize-1)*sizeof(int16)      ); //save history
}

