/****************************************************************************/
/*! \file		fir.h
	\brief		FIR filter
	\author		Iliya Voronov, http://powerdsp.narod.ru
	\version	1.2
     
	FIR filter, 16bit precision.
*/
/*****************************************************************************/
#ifndef _FIR_H_
#define _FIR_H_

#include "comtypes.h"


/******************************************************************************
Internal definitions
******************************************************************************/

/******************************************************************************
Object instance
******************************************************************************/

//! Object instance
typedef struct tagFIR_inst {
  int16           filtSize;     //!< Filter size ( number of tap)
  const int16   * pCoef;        //!< Pointer to coefficients
  /*! history   pBuf[0] - pBuf[filtSize-2]                   , size filtSize-1
      new data  pBuf[filtSize-1] pBuf[filtSize-1+blockSize]  , size blockSize
  */
  int16         * pBuf;         //!< Pointer to buffer 
} tFIR_inst;


/******************************************************************************
External functions
******************************************************************************/

//! Initialize FIR filter object.
/*! Clear buffer and set all parameters. It should be called 
    befor any other functions.
  \param  this          pointer to FIR filter instance
  \param  filtSize      filter size (number of tap)
  \param  pCoef    in   pointer to filter coefficients 
  \param  pBuf     in   pointer to buffer  
*/
_extern_c_ void FIR_init( tFIR_inst     * this, 
                          int16           filtSize, 
                          const int16   * pCoef,
                          int16         * pBuf      );


//! Get pointer to put samples to FIR filter object.
/*! Get pointer to internal object buffer, to write input samples.
    Number of samples be written is blockSize.
  \param	this		pointer to FIR filter  instance
  \return				pointer to write input samples 
*/
_inline_ int16 *  FIR_getPtrToWrite( tFIR_inst  * this   )
{
  return &this->pBuf[this->filtSize-1];
}


//! Process FIR filter calculation for one block.
/*! This function run FIR with the internal data and write output 
    blockSize sampels by the pOut pointer. Function processes 
    blockSize input samples, which should be previousely written 
    by pointer, returned by FIR_getPtrToWrite. 
  \param	this		pointer to FIR filter  instance
  \param	blockSize	number of samples, processed per FIR_procBlock call 
  \param	pOut  out	pointer to write to output samples
*/
_extern_c_ void FIR_procBlock( tFIR_inst    * this,	int16	blockSize,
                               int16        * pOut   );


#endif //_FIR_H_

