/****************************************************************************/
/*! \file		hilb.c
	\brief		Hiltert transform FIR filter
	\author		Iliya Voronov, http://powerdsp.narod.ru
	\version	1.2
     
  Hilbert transform, convert real to complex.
  Magnitude deviation is less than +-0.01 
  in band 300 to 3300 Hz, sample rate 8 kSPS. 
*/
/*****************************************************************************/


#include <string.h>
#include "hilb.h"
#include "baseop.h"


/******************************************************************************
Internal definitions
******************************************************************************/

/*! Hilbert transform FIR filter coeficients.  
	It is antisymmetric FIR filter designed in QED2000 
	files hilb.spc and hilb.flt.
	Magnitude deviation is less than +-0.01 
	in band 300 to 3300 Hz, sample rate 8 kSPS. 
    Number of taps is 33, number of positive coefficients are 8,
    Coefficient taps order:
    C8, 0, C7 ... C2, 0, C1, 0, -C1, 0, -C2 ... -C7, 0, -C8,
    where C1 is first coefficient.
*/
static const int16 HILB_aPosCoef[HILB_POS_COEF_SZ] = 
{
	  20665,  6385,   3281,  1841,   1020,   526,    238,    84
};

/******************************************************************************
External functions
******************************************************************************/

//! Initialize Hilbert object.
/*! Clear buffer and set all parameters. It should be called 
    befor any other functions.
  \param	this			pointer to HILB instance
  \param	pBuf	in		pointer to buffer  
*/
void   HILB_init (	tHILB_inst	* this,
					int16		* pBuf        )
{
	ASSERT( pBuf );
	this->pBuf = pBuf;
}




//! Process Hilbert calculation for one block.
/*! This function run Hilbert transform FIR filter with the internal
    data and write output blockSize of complex sampels by the pOut 
    pointer. Function processes blockSize input samples, which should 
    be previousely written by pointer, returned by HILB_getPtrToWrite. 
  \param    this			pointer to HILB instance
  \param	blockSize		number of samples, processed per HILB_procBlock call 
  \param	pOut	out		pointer to write to output samples
*/
void HILB_procBlock	(	tHILB_inst	* this,  int16	blockSize,
						comp16		* pOut		)
{
	const int16	* pMid; // Middle tap of FIR
	const int16	* pPos; // Pointers to samples for positive coefficients 
	const int16	* pNeg; // Pointers to samples for negative coefficients 
	const int16	* pCoef;
	uint16 			i, j;
	int32			sum1, sum2;

	pMid = &this->pBuf[HILB_POS_COEF_SZ*2-1];
#ifdef COMPILER_C55
	#pragma MUST_ITERATE(1);
#endif //#ifdef COMPILER_C55
	for ( i=0; i < blockSize; i++ ) { // Block process loop 
		pPos  = pMid+1; // Start from middle tap
		pNeg  = pMid-1;
		pCoef = HILB_aPosCoef;
		sum1 = sum2 = 0;

#ifdef COMPILER_C55
		#pragma MUST_ITERATE(1);
#endif //#ifdef COMPILER_C55
		for ( j=0; j < HILB_POS_COEF_SZ; j++ ) { // Antisymmetric FIR loop
			sum1 = _smac( sum1, *pPos, *pCoef);   // Go down with step 2, to leave out zero coefficients 
			pPos += 2;
			sum2 = _smas( sum2, *pNeg, *pCoef++); // Go up with step 2, to leave out zero coefficients
			pNeg -= 2;
		}
		pOut->re = *pMid++; // Delay real part for half FIR size
		pOut->im = hir( _lsadd( sum1, sum2 ) );
		pOut++;
	}
	memcpy(	&this->pBuf[0],
			&this->pBuf[blockSize],
			(HILB_TAP_SZ-1)*sizeof(int16) ); //save history
}



