/****************************************************************************/
/*! \file		hilb.h
	\brief		Hiltert transform FIR filter
	\author		Iliya Voronov, http://powerdsp.narod.ru
	\version	1.2
     
  Hilbert transform, convert real to complex.
  Magnitude deviation is less than +-0.01 
  in band 300 to 3300 Hz, sample rate 8 kSPS 
*/
/*****************************************************************************/

#ifndef _HILB_H_
#define _HILB_H_

#include "comtypes.h"


/******************************************************************************
Internal definitions
******************************************************************************/

//! Hilbert transform FIR number of positive nonzero coefficients
#define HILB_POS_COEF_SZ 8

//! Hilbert transform FIR number of taps
#define HILB_TAP_SZ (HILB_POS_COEF_SZ*4-1)

//! Hilber transform FIR block size
#define HILB_BLOCK_SZ 10


/******************************************************************************
Object instance
******************************************************************************/

//! Object instance
typedef struct tagHILB_inst {
	/*! history   aBuf[0] - aBuf[HILB_TAP_SZ-2]                      , size filtSize-1
		new data  aBuf[HILB_TAP_SZ-1] aBuf[HILB_TAP_SZ-1+blockSize]  , size blockSize
	*/
	int16		* pBuf;		//!< Internal buffer
} tHILB_inst;


/******************************************************************************
External functions
******************************************************************************/

//! Initialize Hilbert object.
/*! Clear buffer and set all parameters. It should be called 
    befor any other functions.
  \param	this			pointer to HILB instance
  \param	pBuf	in		pointer to buffer  
*/
_extern_c_ void HILB_init(	tHILB_inst  * this,
							int16		* pBuf        );


//! Get pointer to put samples to Hilbert object.
/*! Get pointer to internal object buffer, to write real input samples.
    Number of samples be written is blockSize.
  \param    this  pointer to HILB instance
  \return         pointer to write input samples 
*/
_inline_ int16 * HILB_getPtrToWrite( tHILB_inst  * this )
{
	return &this->pBuf[HILB_TAP_SZ-1];
}


//! Process Hilbert calculation for one block.
/*! This function run Hilbert transform FIR filter with the internal
    data and write output blockSize of complex sampels by the pOut 
    pointer. Function processes blockSize input samples, which should 
    be previousely written by pointer, returned by HILB_getPtrToWrite. 
  \param    this			pointer to HILB instance
  \param	blockSize		number of samples, processed per HILB_procBlock call 
  \param    pOut	out		pointer to write output samples
*/
_extern_c_ void HILB_procBlock(	tHILB_inst	* this,  int16	blockSize,
								comp16		* pOut		);


#endif //_HILB_H_

