/****************************************************************************/
/*! \file		resamp.c
	\brief		Resampling interpolator
	\author		Iliya Voronov, http://powerdsp.narod.ru
	\version	1.0
     
	Resample input data from input sample rate to output sample rate.
	Input/output sample rate is 1/RESAMP_SUBFILT_NUMB to RESAMP_SUBFILT_NUMB. 
*/
/*****************************************************************************/

#include <string.h>
#include "resamp.h"
#include "baseop.h"


/******************************************************************************
Internal definitions
******************************************************************************/



//! Number of integer bits in phase value
#define RESAMP_INTEG_BIT_NUM (RESAMP_INTERP_STEP_2PWR+1)


//! Number of fractional bits in phase value
#define RESAMP_FRACT_BIT_NUM (32-RESAMP_INTEG_BIT_NUM)


//! Phase thershold to produce output samples
#define RESAMP_PHASE_THRESH (0x80000000)


//! Phase decrement at put input sample
#define RESAMP_OUT_DECR (1UL<<RESAMP_FRACT_BIT_NUM)



/*! Resampler FIR filter. 
		It is low pass filter defined in resample.spc and resample.flt.
		It is calculated for RESAMP_SUBFILT_NUMB oversampling, 
		at our sampling rate it is all pass filter.
		It is used as RESAMP_SUBFILT_NUMB distinguish FIR filter of 
		(RESAMP_FILT_SIZE+1) size each. Last element is zero, to avoid if in cycle 
    Coefficients order:
		1-st filter 1-st coef, 2-nd coef ... 32-th coef, 0,
		2-nd filter 1-st coef, 2-nd coef ... 32-th coef, 0,
		3-rd filter 1-st coef and so on.
	Filter is low pass, designed in QED2000, see . 
	Sample rate 256 KSPS, pass band to 3300Hz, ripple 0.1 dB; 
	stop band from 4000Hz, ripple 50 dB;
	filter coefficients are scaled and reordered 
	with MultSwap.m procedure.
*/
static const int16 RESAMP_aFiltCoef[RESAMP_SUBFILT_NUMB][RESAMP_FILT_SIZE+1] = {
	{ -41   , -153  , 144   , -169  , 165   , -109  , -11   , 214   , -499  , 864   , -1290 , 1757  , -2230 , 2682  , -3099 , 3586  , 29594 , 2612  , -2682 , 2474  , -2137 , 1735  , -1310 , 907   , -550  , 264   , -55   , -76   , 142   , -156  , 137   , -154  ,  0 },
	{ -41   , -151  , 149   , -181  , 186   , -142  , 32    , 161   , -442  , 814   , -1260 , 1766  , -2306 , 2873  , -3506 , 4596  , 29516 , 1676  , -2258 , 2252  , -2029 , 1699  , -1320 , 942   , -597  , 312   , -97   , -43   , 119   , -142  , 130   , -154  ,  0 },
	{ -40   , -148  , 153   , -191  , 207   , -174  , 76    , 106   , -381  , 756   , -1219 , 1761  , -2366 , 3046  , -3898 , 5638  , 29360 , 782   , -1829 , 2017  , -1907 , 1650  , -1319 , 969   , -638  , 356   , -137  , -10   , 96    , -127  , 122   , -154  ,  0 },
	{ -41   , -145  , 156   , -201  , 226   , -205  , 121   , 49    , -317  , 692   , -1167 , 1742  , -2409 , 3199  , -4272 , 6708  , 29128 , -65   , -1400 , 1772  , -1773 , 1590  , -1307 , 988   , -673  , 397   , -176  , 21    , 72    , -111  , 113   , -152  ,  0 },
	{ -42   , -140  , 158   , -209  , 244   , -235  , 165   , -8    , -248  , 621   , -1106 , 1709  , -2432 , 3330  , -4625 , 7802  , 28820 , -866  , -973  , 1519  , -1627 , 1518  , -1286 , 999   , -703  , 435   , -213  , 53    , 48    , -95   , 103   , -150  ,  0 },
	{ -44   , -135  , 158   , -215  , 260   , -264  , 209   , -67   , -177  , 543   , -1035 , 1661  , -2437 , 3437  , -4954 , 8915  , 28437 , -1616 , -551  , 1259  , -1472 , 1435  , -1254 , 1003  , -726  , 468   , -248  , 84    , 24    , -78   , 92    , -148  ,  0 },
	{ -47   , -129  , 157   , -220  , 274   , -292  , 252   , -126  , -103  , 461   , -954  , 1600  , -2423 , 3520  , -5255 , 10043 , 27982 , -2315 , -138  , 995   , -1307 , 1343  , -1214 , 998   , -744  , 497   , -280  , 113   , 0     , -61   , 81    , -145  ,  0 },
	{ -51   , -122  , 155   , -224  , 287   , -317  , 294   , -186  , -28   , 373   , -864  , 1525  , -2389 , 3576  , -5524 , 11181 , 27457 , -2959 , 262   , 730   , -1136 , 1241  , -1164 , 985   , -756  , 521   , -309  , 141   , -22   , -44   , 69    , -141  ,  0 },
	{ -55   , -115  , 152   , -225  , 297   , -341  , 334   , -244  , 49    , 281   , -766  , 1436  , -2335 , 3606  , -5760 , 12325 , 26864 , -3548 , 650   , 464   , -958  , 1131  , -1106 , 965   , -761  , 542   , -336  , 168   , -45   , -27   , 57    , -137  ,  0 },
	{ -59   , -106  , 147   , -225  , 305   , -362  , 372   , -302  , 127   , 186   , -661  , 1334  , -2261 , 3607  , -5957 , 13469 , 26207 , -4081 , 1022  , 201   , -777  , 1015  , -1041 , 938   , -760  , 557   , -359  , 193   , -67   , -10   , 45    , -132  ,  0 },
	{ -64   , -97   , 141   , -223  , 311   , -380  , 407   , -358  , 205   , 87    , -548  , 1220  , -2168 , 3579  , -6115 , 14610 , 25487 , -4557 , 1375  , -58   , -592  , 892   , -968  , 904   , -754  , 568   , -379  , 215   , -87   , 5     , 32    , -127  ,  0 },
	{ -69   , -88   , 134   , -219  , 315   , -396  , 440   , -412  , 283   , -12   , -429  , 1094  , -2056 , 3522  , -6229 , 15741 , 24708 , -4975 , 1707  , -311  , -407  , 764   , -889  , 863   , -742  , 575   , -396  , 236   , -107  , 22    , 19    , -122  ,  0 },
	{ -75   , -77   , 126   , -214  , 316   , -409  , 470   , -464  , 360   , -115  , -305  , 957   , -1925 , 3436  , -6298 , 16857 , 23875 , -5335 , 2016  , -556  , -221  , 633   , -804  , 816   , -724  , 577   , -410  , 255   , -126  , 37    , 6     , -117  ,  0 },
	{ -80   , -67   , 116   , -206  , 315   , -419  , 497   , -513  , 435   , -218  , -176  , 810   , -1776 , 3319  , -6319 , 17955 , 22991 , -5638 , 2302  , -792  , -37   , 499   , -714  , 764   , -700  , 574   , -420  , 271   , -143  , 53    , -6    , -111  ,  0 },
	{ -86   , -55   , 105   , -197  , 311   , -426  , 520   , -559  , 507   , -320  , -44   , 654   , -1610 , 3173  , -6290 , 19029 , 22060 , -5884 , 2561  , -1017 , 142   , 363   , -620  , 706   , -672  , 567   , -426  , 285   , -159  , 67    , -18   , -104  ,  0 },
	{ -92   , -43   , 94    , -186  , 305   , -429  , 540   , -601  , 577   , -422  , 90    , 490   , -1427 , 2998  , -6209 , 20074 , 21086 , -6074 , 2793  , -1229 , 319   , 226   , -522  , 644   , -638  , 556   , -430  , 296   , -173  , 81    , -31   , -99   ,  0 },
	{ -99   , -31   , 81    , -173  , 296   , -430  , 556   , -638  , 644   , -522  , 226   , 319   , -1229 , 2793  , -6074 , 21086 , 20074 , -6209 , 2998  , -1427 , 490   , 90    , -422  , 577   , -601  , 540   , -429  , 305   , -186  , 94    , -43   , -92   ,  0 },
	{ -104  , -18   , 67    , -159  , 285   , -426  , 567   , -672  , 706   , -620  , 363   , 142   , -1017 , 2561  , -5884 , 22060 , 19029 , -6290 , 3173  , -1610 , 654   , -44   , -320  , 507   , -559  , 520   , -426  , 311   , -197  , 105   , -55   , -86   ,  0 },
	{ -111  , -6    , 53    , -143  , 271   , -420  , 574   , -700  , 764   , -714  , 499   , -37   , -792  , 2302  , -5638 , 22991 , 17955 , -6319 , 3319  , -1776 , 810   , -176  , -218  , 435   , -513  , 497   , -419  , 315   , -206  , 116   , -67   , -80   ,  0 },
	{ -117  , 6     , 37    , -126  , 255   , -410  , 577   , -724  , 816   , -804  , 633   , -221  , -556  , 2016  , -5335 , 23875 , 16857 , -6298 , 3436  , -1925 , 957   , -305  , -115  , 360   , -464  , 470   , -409  , 316   , -214  , 126   , -77   , -75   ,  0 },
	{ -122  , 19    , 22    , -107  , 236   , -396  , 575   , -742  , 863   , -889  , 764   , -407  , -311  , 1707  , -4975 , 24708 , 15741 , -6229 , 3522  , -2056 , 1094  , -429  , -12   , 283   , -412  , 440   , -396  , 315   , -219  , 134   , -88   , -69   ,  0 },
	{ -127  , 32    , 5     , -87   , 215   , -379  , 568   , -754  , 904   , -968  , 892   , -592  , -58   , 1375  , -4557 , 25487 , 14610 , -6115 , 3579  , -2168 , 1220  , -548  , 87    , 205   , -358  , 407   , -380  , 311   , -223  , 141   , -97   , -64   ,  0 },
	{ -132  , 45    , -10   , -67   , 193   , -359  , 557   , -760  , 938   , -1041 , 1015  , -777  , 201   , 1022  , -4081 , 26207 , 13469 , -5957 , 3607  , -2261 , 1334  , -661  , 186   , 127   , -302  , 372   , -362  , 305   , -225  , 147   , -106  , -59   ,  0 },
	{ -137  , 57    , -27   , -45   , 168   , -336  , 542   , -761  , 965   , -1106 , 1131  , -958  , 464   , 650   , -3548 , 26864 , 12325 , -5760 , 3606  , -2335 , 1436  , -766  , 281   , 49    , -244  , 334   , -341  , 297   , -225  , 152   , -115  , -55   ,  0 },
	{ -141  , 69    , -44   , -22   , 141   , -309  , 521   , -756  , 985   , -1164 , 1241  , -1136 , 730   , 262   , -2959 , 27457 , 11181 , -5524 , 3576  , -2389 , 1525  , -864  , 373   , -28   , -186  , 294   , -317  , 287   , -224  , 155   , -122  , -51   ,  0 },
	{ -145  , 81    , -61   , 0     , 113   , -280  , 497   , -744  , 998   , -1214 , 1343  , -1307 , 995   , -138  , -2315 , 27982 , 10043 , -5255 , 3520  , -2423 , 1600  , -954  , 461   , -103  , -126  , 252   , -292  , 274   , -220  , 157   , -129  , -47   ,  0 },
	{ -148  , 92    , -78   , 24    , 84    , -248  , 468   , -726  , 1003  , -1254 , 1435  , -1472 , 1259  , -551  , -1616 , 28437 , 8915  , -4954 , 3437  , -2437 , 1661  , -1035 , 543   , -177  , -67   , 209   , -264  , 260   , -215  , 158   , -135  , -44   ,  0 },
	{ -150  , 103   , -95   , 48    , 53    , -213  , 435   , -703  , 999   , -1286 , 1518  , -1627 , 1519  , -973  , -866  , 28820 , 7802  , -4625 , 3330  , -2432 , 1709  , -1106 , 621   , -248  , -8    , 165   , -235  , 244   , -209  , 158   , -140  , -42   ,  0 },
	{ -152  , 113   , -111  , 72    , 21    , -176  , 397   , -673  , 988   , -1307 , 1590  , -1773 , 1772  , -1400 , -65   , 29128 , 6708  , -4272 , 3199  , -2409 , 1742  , -1167 , 692   , -317  , 49    , 121   , -205  , 226   , -201  , 156   , -145  , -41   ,  0 },
	{ -154  , 122   , -127  , 96    , -10   , -137  , 356   , -638  , 969   , -1319 , 1650  , -1907 , 2017  , -1829 , 782   , 29360 , 5638  , -3898 , 3046  , -2366 , 1761  , -1219 , 756   , -381  , 106   , 76    , -174  , 207   , -191  , 153   , -148  , -40   ,  0 },
	{ -154  , 130   , -142  , 119   , -43   , -97   , 312   , -597  , 942   , -1320 , 1699  , -2029 , 2252  , -2258 , 1676  , 29516 , 4596  , -3506 , 2873  , -2306 , 1766  , -1260 , 814   , -442  , 161   , 32    , -142  , 186   , -181  , 149   , -151  , -41   ,  0 },
	{ -154  , 137   , -156  , 142   , -76   , -55   , 264   , -550  , 907   , -1310 , 1735  , -2137 , 2474  , -2682 , 2612  , 29594 , 3586  , -3099 , 2682  , -2230 , 1757  , -1290 , 864   , -499  , 214   , -11   , -109  , 165   , -169  , 144   , -153  , -41   ,  0 },
};


/******************************************************************************
Forward and inline declarations
******************************************************************************/




/******************************************************************************
External functions
******************************************************************************/

//! Initialize resampler
/*! Clear buffers and set all parameters. It should be called 
    befor any other functions.
  \param  this      pointer to Resampler instance
  \param  inRate    input sample rate
  \param  outRate   output sample rate
*/
void  RESAMP_init(	tRESAMP_inst    * this, 
					uint32 inRate,	uint32 outRate   )
{
	memset( this, 0, sizeof(*this) );
	this->phase = RESAMP_PHASE_THRESH - RESAMP_OUT_DECR;
	RESAMP_setRate( this, inRate, outRate );

#if 0
  { // check subfilter magnitude
    int i, j;
    int32 aSum[RESAMP_SUBFILT_NUMB];
    for( i = 0; i < RESAMP_SUBFILT_NUMB; i++ )
    {
      aSum[i] = 0;
      for( j = 0; j < RESAMP_FILT_SIZE; j++ )
        aSum[i] += RESAMP_aFiltCoef[ j*RESAMP_SUBFILT_NUMB+i ];
    }
  }
#endif

}

//! Set resampler rate.
/*! Set input and output sampling rate, can be called to adjust 
    resampling rate on the fly.
	Ratio = 1/RESAMP_SUBFILT_NUMB to RESAMP_SUBFILT_NUMB
  \param  this      pointer to Resampler instance
  \param  inRate    input sample rate
  \param  outRate   output sample rate
*/
void  RESAMP_setRate(	tRESAMP_inst    * this, 
						uint32 inRate,	uint32 outRate   )
{
	this->outInc = ulfdiv( inRate>>RESAMP_INTEG_BIT_NUM, outRate );
	this->outInc = ulmin( ulmax( this->outInc, 0x00800000UL ), 0x80000000UL ); // < 15.999 && > 0.0625
}




//! Process samples.
/*! This function read input sample, return result code
		and write output sample in place of input sample if output sample is generated.
		If function return result more than on one output sample, it should be called without
		putting input sample untill it returns result one output sample.
		 
    further processing in ENG_dataProc.
  \param  this			pointer to Analog Front End instance
  \param  pSample   pointer to input/output complex sample    
  \return           number of output samples ( 0 to RESAMP_SUBFILT_NUMB ) 
*/
uint16	RESAMP_proc(	tRESAMP_inst	* this, 
						comp16			* pSample )
{
	uint16			i;
	uint16			numb = 0; // output samples number
	uint16			index;	  // samples index
	uint16			subFilt;	// subfilter number (offset in RESAMP_aFiltCoef)
	comp32			sum;
#ifndef _GOST13109_
	const int16		* pCoef;		// pointer to coefficients
#else
	const int32		* pCoef;		// pointer to coefficients
#endif
	const int16		* pRe;			// separate pointer to Re part
	const int16		* pIm;			// separate pointer to Im part

	index	= this->index; // index to put current sample
	this->aBuf[index] = *pSample;
	this->index = (index+1) & (RESAMP_FILT_SIZE-1); // circular increment

	while ( ( this->phase + this->outInc ) <= RESAMP_PHASE_THRESH  ) { 
		this->phase += this->outInc;
		subFilt = (uint16)( (this->phase>>(RESAMP_FRACT_BIT_NUM-RESAMP_INTERP_STEP_2PWR)) & (RESAMP_SUBFILT_NUMB-1) );
		pCoef	= &RESAMP_aFiltCoef[subFilt][0];
		sum.re = sum.im = 0;

#if 1
    // first cycle index to 0
 		pRe	  = &this->aBuf[index].re;
		pIm	  = &this->aBuf[index].im;
#ifdef COMPILER_C55
    #pragma MUST_ITERATE(1)
#endif //#ifdef COMPILER_C55
		for( i = 0; i < (index+1); i++ ) {
			sum.re = _smac( sum.re, *pRe, *pCoef   );
			pRe -= sizeof(comp16)/sizeof(int16);
			sum.im = _smac( sum.im, *pIm, *pCoef++ );
			pIm -= sizeof(comp16)/sizeof(int16);
		}

    // second cycle FILT_SIZE-1 to index+1 plus one extra cycle to avoid if, last coef = 0
		pRe	  = &this->aBuf[RESAMP_FILT_SIZE-1].re;
		pIm	  = &this->aBuf[RESAMP_FILT_SIZE-1].im;
#ifdef COMPILER_C55
    #pragma MUST_ITERATE(1)
#endif //#ifdef COMPILER_C55
		for( i=0; i<(RESAMP_FILT_SIZE-index); i++ ) { // plus one extra cycle to avoid if, last coef = 0
			sum.re = _smac( sum.re, *pRe, *pCoef   );
			pRe -= sizeof(comp16)/sizeof(int16);
			sum.im = _smac( sum.im, *pIm, *pCoef++ );
			pIm -= sizeof(comp16)/sizeof(int16);
		}
#else
		for( i=0; i<RESAMP_FILT_SIZE; i++ )
		{
			sum.re = _smac( sum.re, this->aBuf[index].re, *pCoef );
			sum.im = _smac( sum.im, this->aBuf[index].im, *pCoef );
			index  = (index-1) & (RESAMP_FILT_SIZE-1); // circular decrement, from current to very last
			pCoef++;
		}
#endif
		pSample->re = hir( sum.re );
		pSample->im = hir( sum.im );

		numb++;
		pSample++;
	}

	this->phase -= RESAMP_OUT_DECR;

	ASSERT ( numb <= RESAMP_MAX_RATIO );
	return numb;
}


