/****************************************************************************/
/*! \file		resamp.h
	\brief		Resampling interpolator
	\author		Iliya Voronov, http://powerdsp.narod.ru
	\version	1.0
     
	Resample input data from input sample rate to output sample rate.
	Input/output sample rate is 1/RESAMP_SUBFILT_NUMB to RESAMP_SUBFILT_NUMB. 
*/
/*****************************************************************************/
#ifndef _RESAMP_H_
#define _RESAMP_H_

#include "comtypes.h"

/******************************************************************************
Internal definitions
******************************************************************************/

//! Interpolator step number in 2^N
#define RESAMP_INTERP_STEP_2PWR 5

//! Subfilter number, interpolation step number
#define RESAMP_SUBFILT_NUMB	(1UL<<RESAMP_INTERP_STEP_2PWR)

//! Resampling interpolation filter size
#define RESAMP_FILT_SIZE 32 


//! Resampler maximum ration 
#define RESAMP_MAX_RATIO	(1UL<<RESAMP_INTERP_STEP_2PWR)



/******************************************************************************
Object instance
******************************************************************************/
//! Object instance
typedef struct tagRESAMP_inst {
	uint32		outInc;			//!< Phase increment per output sample, decrement per input sample is 1. = RESAMP_OUT_DECR
	uint32		phase;			//!< Current phase, fractional value with RESAMP_FRACT_BIT_NUM fractional bits, increment for outInc per output sample, decrement for 1. = RESAMP_OUT_DECR per input sample
	uint16		index;			//!< Index in delay line, point to element, next to earliest
	comp16		aBuf[RESAMP_FILT_SIZE];	//!< Circular input samples delay line
} tRESAMP_inst;


/******************************************************************************
External functions
******************************************************************************/


//! Initialize resampler
/*! Clear buffers and set all parameters. It should be called 
    befor any other functions.
  \param  this      pointer to Resampler instance
  \param  inRate    input sample rate
  \param  outRate   output sample rate
*/
_extern_c_ void  RESAMP_init(	tRESAMP_inst    * this, 
								uint32 inRate,	uint32 outRate   );


//! Set resampler rate.
/*! Set input and output sampling rate, can be called to adjus 
    resampling rate on the fly.
	Ratio = 1/RESAMP_SUBFILT_NUMB to RESAMP_SUBFILT_NUMB
  \param  this      pointer to Resampler instance
  \param  inRate    input sample rate
  \param  outRate   output sample rate
*/
_extern_c_ void  RESAMP_setRate(	tRESAMP_inst    * this, 
									uint32 inRate,	uint32 outRate   );



//! Process samples.
/*! This function read input sample, return result code
		and write output sample in place of input sample if output sample is generated.
		If function return result more than on one output sample, it should be called without
		putting input sample untill it returns result one output sample.
		 
    further processing in ENG_dataProc.
  \param  this			pointer to Analog Front End instance
  \param  pSample   pointer to input/output complex sample    
  \return           number of output samples  ( 0 to RESAMP_SUBFILT_NUMB ) 
*/
_extern_c_ uint16		RESAMP_proc(	tRESAMP_inst	* this, 
										comp16			* pSample	);




#endif //_RESAMP_H_
