/****************************************************************************/
/*! \file		bert.c
	\brief		BER tester
	\author		Iliya Voronov, http://powerdsp.narod.ru
	\version	0.0

	Bit error rate tester according to O.153
	Test pattern: 
	1. all 0, all 1
	2. alternating 0/1 with ratio: 1:1, 1:3, 1:7, 3:1, 7:1
	3. pseudo-random test pattern of 511, 2047, 1048575 bits

	Sync established if 8 bits are equal to pattern
	Sync lost if more then BERT_ERR_IN_BURST errors during last 24 bits
*/
/*****************************************************************************/

#include <string.h>
#include <stdio.h>
#include <stdlib.h> 



#include "baseop.h"
#include "bert.h"
#include "fileio.h"




/*****************************************************************************
Internal definitions
*****************************************************************************/


//! Test patterns
static const uint8		BERT_aTestPatt[] = {
	0x00,		//!< BERT_MD_ALL0	-	all 0
	0xff,		//!< BERT_MD_ALL1	-	all 1
	0xaa,		//!< BERT_MD_ALT11	-	1:1
	0xee,		//!< BERT_MD_ALT13	-	1:3
	0xfe,		//!< BERT_MD_ALT17	-	1:7
	0x88,		//!< BERT_MD_ALT31	-	3:1
	0x80		//!< BERT_MD_ALT71	-	7:1
};


//! PRTP step in bits - number of bits, generated per cycle
/*! step is limited by nearest tap - tap1 for 2^20 pattern
*/
#define BERT_PRTP_STEP 2

//! PRTP mask
#define BERT_PRTP_MASK ( ( ( 1UL << BERT_PRTP_STEP ) - 1 ) << ( 24 - BERT_PRTP_STEP ) )

//! PRTP parameters
typedef struct tagBERT_prtpPar {
	uint16		tap1;	//!< 1-st tap
	uint16		tap2;	//!< 2-nd tap and result
} tBERT_prtpPar;

//! PRTP parameters
static const tBERT_prtpPar	BERT_aPrtpPar[] = {
	{	5-BERT_PRTP_STEP,	 9-BERT_PRTP_STEP	},		// 511 bits PRTP
	{	9-BERT_PRTP_STEP,	11-BERT_PRTP_STEP	},		// 2047 bits PRTP
	{	3-BERT_PRTP_STEP,	20-BERT_PRTP_STEP	}		// 1048575 bits PRTP
};


//! Total bits to lost bits max ratio to show BER
#define BERT_TOTAL_TO_LOST 10


/******************************************************************************
Forward and inline declarations
******************************************************************************/

static uint16 BERT_calcTerr( tBERT_inst	* this, uint16	err )
{
	uint16		j, terr;
	// calculate total error
	terr = err;
	for( j = 0; j < (BERT_BURST_LEN_BYTE-1); j++ )
		terr += this->aPrevErr[j];
	// store history of error
	ASSERT( (BERT_BURST_LEN_BYTE-2) > 0 );
	for( j = (BERT_BURST_LEN_BYTE-2); j > 0; j-- )
		this->aPrevErr[j] = this->aPrevErr[j-1];
	this->aPrevErr[0] = err;
	return terr;
}

_inline_ void BERT_clrHist( tBERT_inst	* this )
{
	uint16		j;
	for( j = 0; j < (BERT_BURST_LEN_BYTE-1); j++ )
		this->aPrevErr[j] = 0;
}



/*****************************************************************************
External functions
*****************************************************************************/

//! Init BERT
/*! Init BERT
	\param	this			instance
*/
void BERT_init( tBERT_inst * this )
{
	memset( this, 0, sizeof(*this) );
	this->prg = 0xffffffff;
	this->iFst = true;
}


//! Set mode
/*! Set mode
	\param	this			instance
*/
void BERT_setMode( tBERT_inst * this, eBERT_mode	mode )
{
	ASSERT( mode < BERT_MD_NUM );
	BERT_init( this );
	this->mode = mode;
	if( this->mode <= BERT_MD_ALT_71 )
		this->creg = ( ~BERT_aTestPatt[ this->mode ] ) << 8; // avoid accepting notinitilized value as good data
}




//! Generate test pattern
/*! Generate test pattern
	\param	this			instance
	\param	pData	out		test data
	\param	len				data length 
*/
void BERT_gen(	tBERT_inst	* this,
				uint8		* pData,	uint16	len	)
{
	uint16		i, j;
	uint8		patt;
	uint32		t;
	const tBERT_prtpPar	* pPar;

	if( this->mode <= BERT_MD_ALT_71 ) { // regular patterns
		patt = BERT_aTestPatt[ this->mode ];
		for( i = 0; i < len; i++ )
			*pData++ = patt;
	} else { // pseudo-random patterns
		pPar = &BERT_aPrtpPar[ this->mode - BERT_MD_PR_511 ];
		for( i = 0; i < len; i++ ) {
			for( j = 0; j < 8; j += BERT_PRTP_STEP ) {
				t  = ( this->prg << pPar->tap1 ) ^ ( this->prg << pPar->tap2 );
				this->prg = ( this->prg >> BERT_PRTP_STEP ) | ( t & ( BERT_PRTP_MASK<<8 ) );
			}
			*pData++ = ( this->prg >> ( 32 - 8 - BERT_PRTP_STEP - pPar->tap2 ) ) & 0x00ff;
		}
	}
}


//! Check test pattern
/*! Check test pattern
	\param	this			instance
	\param	pData	in		input data
	\param	pLen	in/out	input/output data length
	\return					filled data length
*/
void BERT_check(	tBERT_inst	* this,
					const uint8	* pData, uint16	len 	)
{
	uint16		i, j, err, terr;
	uint16		patt;
	uint32		t, data;
	const tBERT_prtpPar	* pPar;

	this->res.bit += len*8;

	if( this->mode <= BERT_MD_ALT_71 ) { // regular patterns
		patt = BERT_aTestPatt[ this->mode ];
		for( i = 0; i < len; i++ ) {
			this->creg = ( this->creg >> 8 ) | ( ( *pData++ ) << 8 );
			if( this->iSync ) {
				err = bcnt( ( ( this->creg >> this->shft ) ^ patt ) & 0x00ff );
				terr = BERT_calcTerr( this, err );
				if( terr > BERT_ERR_IN_BURST ) {
					this->iSync = false;
					this->res.lost++;
					this->res.lbit += 8; 
					BERT_clrHist( this );
				} else {
					this->res.err += err;
				}
			} else { // sync lost
				for( j = 0; j < 8; j++ ) // find 
					if( ( ( ( this->creg >> j ) ^ patt ) & 0x00ff ) == 0 ) {
						this->shft = j;
						break;
					}
				if( j < 8 ) {
					this->iSync = true;
					this->iFst = false;
				} else  {
					if( ! this->iFst )
						this->res.lbit += 8;
				}
			}
		}
	} else { // pseudo-random patterns
		pPar = &BERT_aPrtpPar[ this->mode - BERT_MD_PR_511 ];
		for( i = 0; i < len; i++ ) {
			if( this->iSync ) {
				for( j = 0; j < 8; j += BERT_PRTP_STEP ) { // step is limited by nearest tap - tap1 for 2^20 pattern
					t  = ( this->prd << pPar->tap1 ) ^ ( this->prd << pPar->tap2 );
					this->prd = ( this->prd >> BERT_PRTP_STEP ) | ( t & BERT_PRTP_MASK );
				}
				err = bcnt( ( *pData++ ^ ( this->prd >> 16 ) ) & 0x00ff );
				terr = BERT_calcTerr( this, err );
				if( terr > BERT_ERR_IN_BURST ) {
					this->iSync = false;
					this->res.lost++;
					this->res.lbit += 8; 
					BERT_clrHist( this );
				} else {
					this->res.err += err;
				}
			} else { // sync lost
				data = ((uint32)*pData++)<<24;
				this->prd |= data;
				for( j = 0; j < 8; j += BERT_PRTP_STEP ) { // step is limited by nearest tap - tap1 for 2^20 pattern
					t  = ( this->prd << pPar->tap1 ) ^ ( this->prd << pPar->tap2 );
					this->prd >>= BERT_PRTP_STEP;
					data = ( data >> BERT_PRTP_STEP )  ^ ( t & BERT_PRTP_MASK );
				}
				if( ( data & 0x00ff0000 ) == 0 ) {
					this->iSync = true;
					this->iFst = false;
				} else  {
					if( ! this->iFst )
						this->res.lbit += 8;
				}

			}
		}
	}

}


//! Get BER
/*! Get BER in string of 8 character + null,
	if no bits - print empty string,
	if significant number of bits are in sync lost - print text message. 
	\param	this			instance
	\return					BER string in %8.6f format
*/
const char * BERT_getBer( tBERT_inst	* this )
{
	static char	aStr[9];

	if( this->res.bit == 0 ) {
		strcpy( aStr, "        " );
	} else if(  this->res.lbit > ( this->res.bit / BERT_TOTAL_TO_LOST ) ) {
		strcpy( aStr, "synclost" );
	} else {
		snprintf( aStr, 8, "%8.6f", (double)this->res.err/this->res.bit );
		aStr[8] = 0; // if snprintf string is too long there is not terminating character 
	}
	return aStr;
}


//! Print full report for 2 channels
/*! Print full report. 
	\param	this			instance
*/
void BERT_printFull2ch( tBERT_inst	aBert[2] )
{
	FILEIO_printf( "Total bits: %ld+%ld, errors: %ld+%ld, sync losts: %ld+%ld, lost bits: %ld+%ld\n",
		(long)aBert[0].res.bit,		(long)aBert[1].res.bit,
		(long)aBert[0].res.err,		(long)aBert[1].res.err,
		(long)aBert[0].res.lost,	(long)aBert[1].res.lost,
		(long)aBert[0].res.lbit,	(long)aBert[1].res.lbit						);
}






