/****************************************************************************/
/*! \file		bert.h
	\brief		BER tester
	\author		Iliya Voronov, http://powerdsp.narod.ru
	\version	0.0

	Bit error rate tester according to O.153
	Test pattern: 
	1. all 0, all 1
	2. alternating 0/1 with ratio: 1:1, 1:3, 1:7, 3:1, 7:1
	3. pseudo-random test pattern of 511, 2047, 1048575 bits 

	Sync established if 8 bit is correct
	Sync lost if more then BERT_ERR_IN_BURST errors during last 24 bits
*/
/*****************************************************************************/

#ifndef _BERT_H_
#define _BERT_H_

#include "comtypes.h"


/*****************************************************************************
Internal definitions
*****************************************************************************/

//! Test mode
typedef enum enBERT_mode {
	BERT_MD_ALL_0,		//!< all 0
	BERT_MD_ALL_1,		//!< all 1
	BERT_MD_ALT_11,		//!< 1:1
	BERT_MD_ALT_13,		//!< 1:3
	BERT_MD_ALT_17,		//!< 1:7
	BERT_MD_ALT_31,		//!< 3:1
	BERT_MD_ALT_71,		//!< 7:1
	BERT_MD_PR_511,		//!< 2^9  511 bit
	BERT_MD_PR_2047,	//!< 2^11 2047 bit
	BERT_MD_PR_1M,		//!< 2^20 1048575 bit
	BERT_MD_NUM			//!< number of mode
} eBERT_mode;


//! Error burst length in bytes
#define BERT_BURST_LEN_BYTE	3

//! Error number in burst
#define BERT_ERR_IN_BURST 8


//! Error result counters
typedef struct tagBERT_res {
	uint32			bit;	//!< total bits
	uint32			err;	//!< error counter
	uint32			lost;	//!< sync lost counter
	uint32			lbit;	//!< lost bits (not counted becauss of sync lost) counter 
} tBERT_res;



/*****************************************************************************
Object instance
*****************************************************************************/

//! BERT instance
typedef struct tagBERT_inst {
	eBERT_mode		mode;		//!< mode
	bool			iSync;		//!< is sync established in receiver
	bool			iFst;		//!< is first sync established, befor it lost bits are not couned
	uint16			aPrevErr[BERT_BURST_LEN_BYTE-1];	//!< error in previous bytes, receiver

	// regular pattern
	uint16			shft;		//!< shift from pattern to received bits, receiver in pattern mode only
	uint16			creg;		//!< check register, receiver in pattern mode only
	// pseudo-random pattern
	uint32			prg;		//!< pseudo-random test pattern generation register, bit 23-31 , generator
	uint32			prd;		//!< pseudo-random test pattern detection register, bit 16-24, receiver

	tBERT_res		res;		//!< results
} tBERT_inst;


/*****************************************************************************
External functions
*****************************************************************************/

//! Init BERT
/*! Init BERT
	\param	this			instance
*/
_extern_c_ void BERT_init( tBERT_inst * this );


//! Clear BERT
/*! Clear BERT
	\param	this			instance
*/
_inline_ void BERT_clear( tBERT_inst * this )
{
	memset( &this->res, 0, sizeof( this->res ) );
}


//! Set mode
/*! Set mode
	\param	this			instance
*/
_extern_c_ void BERT_setMode( tBERT_inst * this, eBERT_mode	mode );


//! Get error count
/*! Get error and sync lost counters
	\param	this			instance
	\param	pRes	out		pointer to erro results, set if not NULL
	\return					error count
*/
_inline_ uint32 BERT_getErr( tBERT_inst	* this,	tBERT_res	* pRes )
{
	if( pRes )
		*pRes  = this->res;
	return this->res.err;
}



//! Generate test pattern
/*! Generate test pattern
	\param	this			instance
	\param	pData	out		test data
	\param	len				data length 
*/
_extern_c_ void BERT_gen(	tBERT_inst	* this,
							uint8		* pData,	uint16	len	);



//! Check test pattern
/*! Check test pattern
	\param	this			instance
	\param	pData	in		input data
	\param	pLen	in/out	input/output data length
	\return					filled data length
*/
_extern_c_ void BERT_check(	tBERT_inst	* this,
							const uint8	* pData, uint16	len 	);



//! Get BER
/*! Get BER in string of 8 character + null,
	if no bits - print empty string,
	if significant number of bits are in sync lost - print text message. 
	\param	this			instance
	\return					BER string in %8.6f format
*/
_extern_c_ const char * BERT_getBer( tBERT_inst	* this );


//! Print full report for 2 channels
/*! Print full report. 
	\param	this			instance
*/
_extern_c_ void BERT_printFull2ch( tBERT_inst	aBert[2] );


#endif //#ifndef _BERT_H_



