/****************************************************************************/
/*! \file		fileIo.c
	\brief		File IO operations
	\author		Iliya Voronov, http://powerdsp.narod.ru
	\version	1.0
  
	Functionality:
	Write text message to log file,
	Write modem samples to 16-bit stereo PCM file
\*/
/*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
//#include <math.h>

#include "comtypes.h"
#include "baseop.h"
#include "fileIo.h"


/*****************************************************************************
Internal definitions
*****************************************************************************/

//! Text log files, copy of std out, initalized in FILEIO_init
static FILE * FILEIO_pLogFile = NULL;


//! Wave file format header
typedef struct tagFILEIO_waveHdr {
	uint32		chunkId;
	uint32		chunkSz;
	uint32		format;
	uint32		subch1Id;
	uint32		subch1Sz;
#ifndef COMPILER_C55
	uint16		audForm;
	uint16		numCh;
#else //#ifndef COMPILER_C55
	uint16		numCh;			// swapped to simplify big-endian to little endian conversion
	uint16		audForm;		// swapped to simplify big-endian to little endian conversion
#endif //#ifndef COMPILER_C55
	uint32		sampRate;
	uint32		byteRate;
#ifndef COMPILER_C55
	uint16		blckAlig;
	uint16		bitPerSamp;
#else //#ifndef COMPILER_C55
	uint16		bitPerSamp;		// swapped to simplify big-endian to little endian conversion
	uint16		blckAlig;		// swapped to simplify big-endian to little endian conversion
#endif //#ifndef COMPILER_C55
	uint32		subch2Id;
	uint32		subch2Sz;
} tFILEIO_waveHdr;

//! Wave file default header
static const tFILEIO_waveHdr FILEIO_defWaveHdr = {
	0x46464952,	// "RIFF" - little endian
	36,			// 36 + data size in byte
	0x45564157,	// "WAVE"
	0x20746d66,	// "fmt "
	16,			// subch1 size
#ifndef COMPILER_C55
	1,			// PCM
	2,			// stereo
#else //#ifndef COMPILER_C55
	2,			// stereo					// swapped to simplify big-endian to little endian conversion
	1,			// PCM						// swapped to simplify big-endian to little endian conversion
#endif //#ifndef COMPILER_C55
	8000,		// 8000 SPS
	32000,		// 8000 SPS * 2 stereo * 16 bitsamp /8
#ifndef COMPILER_C55
	4,			// 2 stereo * 16 bitsamp/8
	16,			// 16 bit samples
#else //#ifndef COMPILER_C55
	16,			// 16 bit samples			// swapped to simplify big-endian to little endian conversion
	4,			// 2 stereo * 16 bitsamp/8	// swapped to simplify big-endian to little endian conversion
#endif //#ifndef COMPILER_C55
	0x61746164,	// "data"
	0,			// subch2 size = data size in byte
};

//! Save samples file names
static const char	* FILEIO_apSampFname[] = {
	"samp0.wav", "samp1.wav"
};

//! Save bits file names
static const char	* FILEIO_apBitsFname[] = {
	"bitsTr0.dat", "bitsTr1.dat", "bitsRc0.dat", "bitsRc1.dat"
};


//! Internal block size 
#define FILEIO_BL_SZ 32


/******************************************************************************
Forward and inline declarations
******************************************************************************/

//! Platform independent fread function to read 16-bit values
static size_t FILEIO_fread16( int16 * pBuf, size_t count, FILE * pFile );

//! Platform independent fwrite function to write 16-bit values
static size_t FILEIO_fwrite16( const int16 * pBuf, size_t count, FILE * pFile );

//! Convert to little-endian format Swap 16-bit halves in 32-bit values for TMS320C54/55
static void FILEIO_toLittleEnd( int32 * pBuf, size_t sz );



//! Put WAVE file format header
_inline_ bool FILEIO_putWavHdr( FILE * pFile ) {
	tFILEIO_waveHdr	waveHdr = FILEIO_defWaveHdr;
	size_t			sz =  sizeof( tFILEIO_waveHdr )/sizeof(int16);

	FILEIO_toLittleEnd( (int32*)&waveHdr, sz/2 );
	sz = FILEIO_fwrite16( (int16*)&waveHdr, sz, pFile );
	return ( sz == sizeof( tFILEIO_waveHdr )/sizeof(int16) );
}

//! Set data size in wave file header 
static void FILEIO_setWavSz( FILE * pFile ) {
	fpos_t			fpos;
	tFILEIO_waveHdr	waveHdr = FILEIO_defWaveHdr;
	size_t			sz =  sizeof( tFILEIO_waveHdr )/sizeof(int16);
	size_t			hdrSz = sizeof( tFILEIO_waveHdr )/sizeof(int16)*2;

	fgetpos( pFile, &fpos );
	waveHdr.subch2Sz = (uint32)( fpos - hdrSz ); // data size
	waveHdr.chunkSz = waveHdr.subch2Sz + 36;
	rewind( pFile );

	FILEIO_toLittleEnd( (int32*)&waveHdr, sz/2 );
	FILEIO_fwrite16( (int16*)&waveHdr, sz, pFile );
}



/******************************************************************************
External functions
******************************************************************************/

//! Initialize File IO object.
/*! Initialize File IO object.
	\param	this			pointer to File IO instance
*/
void  FILEIO_init( tFILEIO_inst * this   )
{
	memset( this, 0, sizeof(*this) );
}


//! Configure File IO object.
/*! This function configure object according to the configuration
    settings got from parsed configuration file. It opens all 
    required files.
	\param	this			pointer to File IO instance
	\param	pCfg	in		pointer to parser result 
*/
void  FILEIO_config       ( tFILEIO_inst      * this, 
                            const tPARSE_cfg  * pCfg )
{
	uint16		i;
	ASSERT( pCfg );

	this->logPerMs = pCfg->logPerMs;
	for( i = 0; i < PARSE_SF_NUM; i++ )
		if( pCfg->aiSaveSamp[i] ) {
			this->apSamp[i] = fopen( FILEIO_apSampFname[ i ], "wb" );
			if( this->apSamp[i] == NULL ) {
				FILEIO_printf( "Can not open file %s for wb\n", FILEIO_apSampFname[ i ] );
			} else {
				if( ! FILEIO_putWavHdr( this->apSamp[i] ) )
					this->apSamp[i] = NULL;
			}
		} else {
			this->apSamp[i] = NULL;
		}
	for( i = 0; i < PARSE_BF_NUM; i++ )
		if( pCfg->aiSaveBits[i] ) {
			this->apBits[i] = fopen( FILEIO_apBitsFname[ i ], "wt" );
			if( this->apBits[i] == NULL )
				FILEIO_printf( "Can not open file %s for wt\n", FILEIO_apBitsFname[ i ] );
		} else {
			this->apBits[i] = NULL;
		}
}


//! Write modem samples to file.
/*! Write modem samples to stereo 16-bit PCM file,
	Tr samples to left channel,
	Rc samples to right channel.
	\param	this			pointer to File IO instance
	\param	num				modem number 0/1
	\param	pTr		in		pointer to Tr samples
	\param	pRc		in		pointer to Rc samples
	\param	len				samples lenght
	\return					true if successfully written          
*/
bool FILEIO_wrSamp	(	tFILEIO_inst	* this,	uint16			  num,	
						const int16		* pTr,	const int16		* pRc,	
						uint16			  len							 )

{
	uint16		part, i, cntr;
	int16		aBuf[FILEIO_BL_SZ*2];
	int16	  *	ptr;
	ASSERT( num < PARSE_SF_NUM );

	if( ! this->apSamp[num] )
		return false;

	cntr = len * 2; // count from len down to 0
	while( len > 0 ) {
		part = umin( len, FILEIO_BL_SZ );
		ptr = aBuf;
		for( i = 0; i < part; i++ ) {
			*ptr++ = *pTr++;
			*ptr++ = *pRc++;
		}
		cntr -= FILEIO_fwrite16( aBuf, part*2, this->apSamp[num] );
		len -= part;
	}

	return ( cntr == 0 );
}


//! Write modem bits to file.
/*! Write modem bits to text file.
	\param	this			pointer to File IO instance
	\param	num				bits file number 0-Tr0, 1-Tr1, 2-Rc0, 3-Rc1
	\param	pData	in		pointer to Rc bytes
	\param	len				bytes lenght
*/
void FILEIO_wrBits	(	tFILEIO_inst	* this,		ePARSE_bitsFile	num,
						const uint8		* pData,	uint16			len		)
{
	ASSERT( num < PARSE_BF_NUM );
	if( ! this->apBits[num] )
		return;

	while( len-- > 0 )
		fprintf( this->apBits[num], "%02x\n", (int)(*pData++)&0x00ff  );
}



//! Flush all output log files.
/*! This function flushes all output log files to 
    allow see results before testbench operation complete.
	\param	this          pointer to File IO instance
*/
void  FILEIO_flush ( tFILEIO_inst * this )
{
	uint16 i;
	for ( i = 0; i < PARSE_SF_NUM; i++ )
		if ( this->apSamp[i] )
			fflush( this->apSamp[i] );
	for ( i = 0; i < PARSE_BF_NUM; i++ )
		if ( this->apBits[i] )
			fflush( this->apBits[i] );
}



//! Close all open files.
/*! This function close all input samples files and 
    output log files of File IO object. It is used to 
    finilize current session before opening other one.
    Text message file is not closed.
  \param    this          pointer to File IO instance
*/
void  FILEIO_close ( tFILEIO_inst * this )
{
	uint16 i;
	for ( i = 0; i < PARSE_SF_NUM; i++ )
		if ( this->apSamp[i] ) {
			FILEIO_setWavSz( this->apSamp[i] );
			fclose( this->apSamp[i] );
		}
	for ( i = 0; i < PARSE_BF_NUM; i++ )
		if ( this->apBits[i] )
			fclose( this->apBits[i] );
}




//! Open log file for write.
/*! Open log file for write.
	\param	pFname	in		pointer to file name
*/
void  FILEIO_openLog  ( const char * pFname  )
{
	FILEIO_pLogFile = fopen( pFname, "wt" );
	if ( FILEIO_pLogFile ) {
		FILEIO_printf( "File %s open to write log\n", pFname );
	} else { // can not use write to message log file in FILEIO_printf 
		printf( "Can not open file %s for wt\n", pFname );
	}
}


//! Print text message.
/*! This function prints message to stdout and to text log file.
	\param	pFormat	in		pointer to format strinf
	\param	...		in		other parameters
*/
void  FILEIO_printf         ( const char * pFormat,  ...  )
{
	va_list arg_list;

	va_start( arg_list, pFormat);
	vprintf( pFormat, arg_list);
	va_end( arg_list);

	if ( FILEIO_pLogFile ) {
		va_start( arg_list, pFormat);
		vfprintf( FILEIO_pLogFile, pFormat, arg_list );
		va_end( arg_list);
		fflush( FILEIO_pLogFile );
	}
}




/******************************************************************************
Internal functions
******************************************************************************/


/*! Platform independent fread function to read 16-bit values
	This function fixes TMS320C54/55 problem with 16 bit byte
*/
static size_t FILEIO_fread16( int16 * pBuf, size_t count, FILE * pFile ) 
{
#ifndef COMPILER_C55
	return fread( pBuf, sizeof(int16), count, pFile );
#else //#ifndef COMPILER_C55
	size_t	i, sz = 0;
	uint8	aByte[2]; // 2 bytes to keep int16
	for( i = 0; i < count; i++ ) {
		sz += fread( aByte, 1, 2, pFile );
		*pBuf++ = (aByte[0]&0xFF) | (aByte[1]<<8);
	}
	return sz>>1;
#endif //#ifndef COMPILER_C55
}

/*! Platform independent fwrite function to write 16-bit values
	This function fixes TMS320C54/55 problem with 16 bit byte
*/
static size_t FILEIO_fwrite16( const int16 * pBuf, size_t count, FILE * pFile )
{
#ifndef COMPILER_C55
	return fwrite( pBuf, sizeof(int16), count, pFile );
#else //#ifndef COMPILER_C55
	size_t	i, sz = 0;
	int16	t;
	uint8	aByte[2]; // 2 bytes to keep int16
	for( i = 0; i < count; i++ ) {
		t = *pBuf++;
		aByte[0] = (uint8)(t&0xFF);
		aByte[1] = (uint8)(t>>8);
		sz += fwrite( aByte, 1, 2, pFile );
	}
	return sz>>1;
#endif //#ifndef COMPILER_C55
}


/*! Convert to little-endian format
	Swap 16-bit halves in 32-bit values for TMS320C54/55
*/
static void FILEIO_toLittleEnd( int32 * pBuf, size_t count )
{
#ifndef COMPILER_C55
#else //#ifndef COMPILER_C55
	while( count-- > 0 ) {
		*pBuf = ulswap( *pBuf );
		pBuf++;
	}
#endif //#ifndef COMPILER_C55
}
