/****************************************************************************/
/*! \file		fileIo.h
	\brief		File IO operations
	\author		Iliya Voronov, http://powerdsp.narod.ru
	\version	1.0
  
	Functionality:
	Write text message to log file,
	Write modem samples to 16-bit stereo PCM file
*/
/*****************************************************************************/

#ifndef _FILE_IO_H_
#define _FILE_IO_H_

#include <stdio.h>
#include "comtypes.h"
#include "parse.h"


/******************************************************************************
Internal definitions
******************************************************************************/




/******************************************************************************
Object instance
******************************************************************************/


//! File IO instance
typedef struct tagFILEIO_inst {
	FILE		* apSamp[PARSE_SF_NUM];	//!< pointers to files to save modem 0/1 TrRc samples to stereo file
	FILE		* apBits[PARSE_BF_NUM];	//!< pointers to files to save modem Tr0/Tr1/Rc0/Rc1 bits
	uint32		  logPerMs;		//!< logging period in ms
} tFILEIO_inst;



/******************************************************************************
External functions
******************************************************************************/

//! Initialize File IO object.
/*! Initialize File IO object.
	\param	this			pointer to File IO instance
*/
_extern_c_  void  FILEIO_init		(	tFILEIO_inst	* this   );


//! Configure File IO object.
/*! This function configure object according to the configuration
    settings got from parsed configuration file. It opens all 
    required files.
	\param	this			pointer to File IO instance
	\param	pCfg	in		pointer to parser result 
*/
_extern_c_  void  FILEIO_config		(	tFILEIO_inst		* this, 
										const tPARSE_cfg	* pCfg );


//! Write modem samples to file.
/*! Write modem samples to stereo 16-bit PCM file,
	Tr samples to left channel,
	Rc samples to right channel.
	\param	this			pointer to File IO instance
	\param	num				modem number 0/1
	\param	pTr		in		pointer to Tr samples
	\param	pRc		in		pointer to Rc samples
	\param	len				samples lenght
	\return					true if successfully written          
*/
_extern_c_  bool FILEIO_wrSamp	(	tFILEIO_inst	* this,	uint16			  num,	
									const int16		* pTr,	const int16		* pRc,	
									uint16			  len         );


//! Write modem bits to file.
/*! Write modem bits to text file.
	\param	this			pointer to File IO instance
	\param	num				bits file number 0-Tr0, 1-Tr1, 2-Rc0, 3-Rc1
	\param	pData	in		pointer to Rc bytes
	\param	len				bytes lenght
	\return					true if successfully written          
*/
_extern_c_  void FILEIO_wrBits	(	tFILEIO_inst	* this,		ePARSE_bitsFile	num,	
									const uint8		* pData,	uint16			len		);


//! Flush all output log files.
/*! This function flushes all output log files to 
    allow see results before testbench operation complete.
	\param	this          pointer to File IO instance
*/
_extern_c_ void  FILEIO_flush ( tFILEIO_inst * this );


//! Close all open file.
/*! This function close all input samples files and 
    output log files of File IO object. It is used to 
    finilize current session before opening other one.
    Text message file is not closed.
	\param	this			pointer to File IO instance
*/
_extern_c_ void  FILEIO_close ( tFILEIO_inst * this );


//! Open log file for write.
/*! Open log file for write.
	\param	pFname	in		pointer to file name
*/
_extern_c_  void  FILEIO_openLog ( const char * pFname  );


//! Print text message.
/*! This function prints message to stdout and to text log file.
	\param	pFormat	in		pointer to format strinf
	\param	...		in		other parameters
*/
_extern_c_  void  FILEIO_printf	( const char * pFormat,  ...  );




#endif // _FILE_IO_H_
