/****************************************************************************/
/*! \file		mdm.c
	\brief		Modem wrapper
	\author		Iliya Voronov, http://powerdsp.narod.ru
	\version	0.0

	Wrapper to unify modem datapump interface.
	Modem put/get data by callback.
*/
/*****************************************************************************/

#include <string.h>

#include "mdm.h"





/*****************************************************************************
Internal definitions
*****************************************************************************/


/******************************************************************************
Forward and inline declarations
******************************************************************************/



/*****************************************************************************
External definitions
*****************************************************************************/

//! Data rate enum to bits per second
const uint16 MDM_aDrate2bps[ MDM_DR_NUM ] = {
	0,		//!< MDM_DR_OFF	   
	1200,	//!< MDM_DR_1200
	2400,	//!< MDM_DR_2400
};

//! Default configuration 
const tMDM_cfg MDM_defCfg = {
	false,			//!< iCall	- is callind side
	MDM_DRMSK_1200 | MDM_DRMSK_2400	//!< rtMsk	- allowed data rate mask
};



static int get_bit_func( void * user_data )
{
	tMDM_inst	  * this = (tMDM_inst*)user_data;
	tMDM_dbuf	  * pBuf = &this->trBuf;

	if( pBuf->bitCntr > 0 ) {
		pBuf->bitCntr--;
		pBuf->aBuf[pBuf->byteCntr] >>= 1;
	} else {
		pBuf->byteCntr++;
		if( pBuf->byteCntr == pBuf->grSz ) {
			pBuf->rqSz = MDM_DBUF_SZ;
			this->evtHndl.fCb( this->evtHndl.pHst, MDM_ET_GET_TR, pBuf );
			ASSERT( pBuf->grSz > 0 );
			ASSERT( pBuf->grSz <= pBuf->rqSz );
			pBuf->byteCntr = 0;
		}
		pBuf->bitCntr = MDM_BITS_PER_ABUF0 - 1;
	}
	return pBuf->aBuf[pBuf->byteCntr] & 0x0001;
}

static void put_bit_func(void *user_data, int bit)
{
	tMDM_inst	  * this = (tMDM_inst*)user_data;
	tMDM_dbuf	  * pBuf = &this->rcBuf;

	if( pBuf->bitCntr < MDM_BITS_PER_ABUF0 ) {
		pBuf->bitCntr++;
		pBuf->aBuf[pBuf->byteCntr] >>= 1;
	} else {
		pBuf->byteCntr++;
		if( pBuf->byteCntr == MDM_DBUF_SZ ) {
			pBuf->rqSz = MDM_DBUF_SZ;
			this->evtHndl.fCb( this->evtHndl.pHst, MDM_ET_PUT_RC, pBuf );
			ASSERT( pBuf->grSz > 0 );
			ASSERT( pBuf->grSz <= pBuf->rqSz );
			pBuf->byteCntr -= pBuf->grSz;
			memmove( &pBuf->aBuf[0], &pBuf->aBuf[pBuf->grSz], pBuf->byteCntr*sizeof(pBuf->aBuf[0]) );
		}
		pBuf->aBuf[pBuf->byteCntr] = 0;
		pBuf->bitCntr = 1;
	}
	pBuf->aBuf[pBuf->byteCntr] |= (bit & 0x0001) << 7;
}



/*****************************************************************************
External functions
*****************************************************************************/

//! Init modem
/*! Init modem
	\param	this			instance
	\return					true if success
*/
bool		MDM_init(		tMDM_inst * this,	const tMDM_evtHndl	* pEvtHndl )
{
	void *	ptr;

	memset( this, 0, sizeof(*this) );
	ASSERT( pEvtHndl );
	ASSERT( pEvtHndl->fCb );
	this->evtHndl = *pEvtHndl;
	this->cfg = MDM_defCfg;

	this->trBuf.byteCntr = (uint16)-1;

	ptr = fsk_tx_init(	&this->tr,		&preset_fsk_specs[FSK_V23CH1],
						get_bit_func,	 this							);
	if( ! ptr )
		return false;
	ptr = fsk_rx_init(	&this->rc,		&preset_fsk_specs[FSK_V23CH1],
						1,				
						put_bit_func,	this							);
	if( ! ptr )
		return false;

	//fsk_tx_power( &this->tr, -80. );

	return true;
}


//! Configure modem
/*! Configure modem
	\param	this			instance
*/
void		MDM_config(		tMDM_inst * this,	const tMDM_cfg	* pCfg )
{
	ASSERT( pCfg );
	this->cfg = *pCfg;

//	fsk_tx_power( &this->tr, -6. );
}


//! Get modem state
/*! Get modem state
	\param	this			instance
*/
tMDM_stat *	MDM_getStat(	tMDM_inst * this )
{
	return &this->stat;
}


//! Control modem
/*! Disconnect modem and connect again
	\param	this			instance
	\param	cmd				control command
	\param	pArg	in/out	optional command argument
	\return					true if success
*/
bool	MDM_ctrl(	tMDM_inst	* this, eMDM_ctrlCmd cmd,
					void		* pArg						)
{
	return true;
}




//! Modulate and demodulate samples.
/*! Modulate and demodulate samples.
	Frame size is MDM_FR_SZ.
	Modem gets transmitting data by MDM_ET_GET_TR callback
	and puts received data by MDM_ET_PUT_RC callback.
	\param	this			instance
	\param	pTr	out			transmitter output samples
	\param	pRc	in			receiver input samples
*/
void	MDM_proc(	tMDM_inst	* this,	
					int16		* pTr,	const int16	* pRc		)
{
	uint16 len;
//	return V23_mod( this, pData, len, pSamp );
	len = fsk_tx( &this->tr, pTr, MDM_FR_SZ );
	ASSERT( len == MDM_FR_SZ );
	len = fsk_rx( &this->rc, pRc, MDM_FR_SZ );
	ASSERT( len == 0 );
}







