/****************************************************************************/
/*! \file		mdm.h
	\brief		Modem wrapper
	\author		Iliya Voronov, http://powerdsp.narod.ru
	\version	0.0

	Wrapper to unify modem datapump interface.
	Modem put/get data by callback.
	TODO: implement MDM_ctrl
*/
/*****************************************************************************/

#ifndef _MDM_H_
#define _MDM_H_

#include "comtypes.h"

#include "inttypes.h"
#include "telephony.h"
#include "complex.h"
#include "dds.h"
#include "async.h"
#include "power_meter.h"
#include "fsk.h"


/*****************************************************************************
Internal definitions
*****************************************************************************/

//! Sample rate
#define MDM_SAMP_RATE 8000

//! Sample frame in ms
#define MDM_FR_MS 10

//! Sample frame size
#define MDM_FR_SZ (MDM_SAMP_RATE/1000*MDM_FR_MS)



//! Data rate 
typedef enum enMDM_drate {
	MDM_DR_OFF,			//!< not yet connected
	MDM_DR_1200,		//!< 1200
	MDM_DR_2400,		//!< 2400
	MDM_DR_NUM			//!< number of rates
} eMDM_drate;

//! Data rate mask
typedef enum enMDM_drMsk {
	MDM_DRMSK_OFF	= 0,
	MDM_DRMSK_1200	= (1<<MDM_DR_1200),
	MDM_DRMSK_2400	= (1<<MDM_DR_2400)
} eMDM_drMsk;




//! Data buffer size
#define MDM_DBUF_SZ 5

//! Data buffer and put/get argument of call back function
typedef struct tagMDM_dbuf {
	uint16		rqSz;		//!< requested data size in bytes,		PUT_RC - given		GET_TR - requested
	uint16		grSz;		//!< granted data size in bytes,	PUT_RC - accepted	GET_TR - filled
	uint8		aBuf[MDM_DBUF_SZ];	//!< buffer
	uint16		bitCntr;	//!< bit counter
	uint16		byteCntr;	//!< byte counter
} tMDM_dbuf;

//! Bits per aBuf[0]
#define MDM_BITS_PER_ABUF0 8




//! Event type
typedef enum enMDM_evtType {
	MDM_ET_CHNG_ST,		//!< change modem state
	MDM_ET_PUT_RC,		//!< put received data to framwork 
	MDM_ET_GET_TR,		//!< get transmitting data from framework
	MDM_ET_NUM			//!< number of rates
} eMDM_evtType;

//! Call back function
typedef void (* fMDM_callBack)( void * pHst, eMDM_evtType evt, const void * pArg );

//! Event handel structure
typedef struct tagMDM_evtHndl {
	fMDM_callBack	  fCb;		//!< call back function
	void			* pHst;		//!< called host
} tMDM_evtHndl;




//! Modem configuration
typedef struct tagMDM_cfg {
	bool		iCall;			//!< is callind side
	uint16		rtMsk;			//!< allowed data rate mask
} tMDM_cfg;

//! Modem state
typedef struct tagMDM_stat {
	bool		iCall;			//!< is callind side
	bool		iCD;			//!< is carier detected 
	bool		iCTS;			//!< is CTS
	eMDM_drate	MDM_RT_1200;	//!< data rate (if CTS true)
} tMDM_stat;



//! Control command
typedef enum enMDM_ctrlCmd {
	MDM_CC_CONN,				//!< connect
	MDM_CC_DISCON,				//!< disconnect
	MDM_CC_SET_CFG,				//!< set configuration
	MDM_CC_GET_STAT				//!< get state
} eMDM_ctrlCmd;


/*****************************************************************************
External declarations
*****************************************************************************/

//! Data rate enum to bits per second
extern const uint16 MDM_aDrate2bps[ MDM_DR_NUM ];

//! Default configuration 
extern const tMDM_cfg MDM_defCfg;



/*****************************************************************************
Object instance
*****************************************************************************/

//! Modem instance
typedef struct tagMDM_inst {
	tMDM_evtHndl	evtHndl;	//!< event handle
	tMDM_cfg		cfg;		//!< configuration
	tMDM_stat		stat;		//!< current state
	tMDM_dbuf		trBuf;		//!< transmit data buffer
	tMDM_dbuf		rcBuf;		//!< receive data buffer

	fsk_tx_state_t	tr;
	fsk_rx_state_t	rc;

} tMDM_inst;


/*****************************************************************************
External functions
*****************************************************************************/

//! Init modem
/*! Init modem
	\param	this			instance
	\return					true if success
*/
_extern_c_ bool	MDM_init(	tMDM_inst * this,	const tMDM_evtHndl	* pEvtHndl );


//! Configure modem
/*! Configure modem
	\param	this			instance
*/
_extern_c_ void	MDM_config(	tMDM_inst * this,	const tMDM_cfg	* pCfg );


//! Get modem state
/*! Get modem state
	\param	this			instance
*/
_extern_c_ tMDM_stat *	MDM_getStat(	tMDM_inst * this );


//! Control modem
/*! Disconnect modem and connect again
	\param	this			instance
	\param	cmd				control command
	\param	pArg	in/out	optional command argument
	\return					true if success
*/
_extern_c_ bool	MDM_ctrl(	tMDM_inst	* this, eMDM_ctrlCmd cmd,
							void		* pArg						);




//! Modulate and demodulate samples.
/*! Modulate and demodulate samples.
	Frame size is MDM_FR_SZ.
	Modem gets transmitting data by MDM_ET_GET_TR callback
	and puts received data by MDM_ET_PUT_RC callback.
	\param	this			instance
	\param	pTr	out			transmitter output samples
	\param	pRc	in			receiver input samples
*/
_extern_c_ void	MDM_proc(	tMDM_inst	* this,	
							int16		* pTr,	const int16	* pRc	);



#endif //#ifndef _MDM_H_



