/****************************************************************************/
/*! \file		parse.c
	\brief		Parser for configuration file
	\author		Iliya Voronov, http://powerdsp.narod.ru
	\version	1.0
  
	It is parser for configuration file. Test bench commands are parsed here,
	if command is not valide test bench command, it is transfered to 
	channel simulator for parsing.
	TODO: add BERT mode setting and script termination commands
*/
/*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "parse.h"
#include "chsim.h"
#include "fileio.h"

#define PARSE_STR_SZ 256

/******************************************************************************
Internal data structures
******************************************************************************/

//! Command processing function for particular type of command
/*! \param	this		pointer to parser inctanse
    \param	pArg	in	pointer to arguments string of command
	\param	val			optional argument value from command description
	\return				true on success
*/
typedef bool (* fPARSE_proc)( tPARSE_inst * this, char * pArg, int val );


//! Command description structure
typedef struct tagPARSE_cmdDesc {
	char		  * pName;		//!< command name
	fPARSE_proc		fProc;		//!< command processing functions
	bool			iEnd;		//!< is end of session
	int				val;		//!< optional command value
} tPARSE_cmdDesc;




/******************************************************************************
Forward declaration
******************************************************************************/

//! Print out parameters of parsed session
static void PARSE_printCfg( const tPARSE_cfg * pCfg );


//! Command processing function - Save TrRc samples
static bool PARSE_procSaveSamp( tPARSE_inst * this, char * pArg, int val );

//! Command processing function - Save bits
static bool PARSE_procSaveBits( tPARSE_inst * this, char * pArg, int val );

//! Command processing function - Set logging period
static bool PARSE_procLogPer( tPARSE_inst * this, char * pArg, int val );

//! Command processing function - Run session
static bool PARSE_procRun( tPARSE_inst * this, char * pArg, int val );



/******************************************************************************
Internal data
******************************************************************************/
//! White characters
static const char   * pWhite = " \t\n";



//! Command description
static const tPARSE_cmdDesc PARSE_aCmdDesc[] = {
	{ "savesamp0",		PARSE_procSaveSamp,	false,	0				},	//!< Save TrRc0 samples to stereo file
	{ "savesamp1",		PARSE_procSaveSamp,	false,	1				},	//!< Save TrRc1 samples to stereo file
	{ "savetrbits0",	PARSE_procSaveBits,	false,	PARSE_BF_TR0	},	//!< Save Tr0 bits to text file
	{ "savetrbits1",	PARSE_procSaveBits,	false,	PARSE_BF_TR1	},	//!< Save Tr1 bits to text file
	{ "savercbits0",	PARSE_procSaveBits,	false,	PARSE_BF_RC0	},	//!< Save Rc0 bits to text file
	{ "savercbits1",	PARSE_procSaveBits,	false,	PARSE_BF_RC1	},	//!< Save Rc1 bits to text file
	{ "logper",			PARSE_procLogPer,	false					},	//!< Set logging period in ms
	{ "runms",			PARSE_procRun,		true,	0				},	//!< Run session, last command
	{ "runbits",		PARSE_procRun,		true,	1				},	//!< Run session, last command
	{ NULL															}	//!< End of list          
};



//! Default session configuration
static const tPARSE_cfg	PARSE_defCfg = {
	{ false, false },				// aiSaveSamp	- is Save modem 0/1 TrRc samples to stereo file
	{ false, false, false, false },	// aiSaveBits	- is Save modem Tr0/Tr1/Rc0/Rc1 bits to file
	1000,		// logPerMs		- Set logging period in ms
	false,		// iLenInBits	- is session length in bits, otherwise in ms
	10000		// sesLen		- session length in ms or in bits
};


/******************************************************************************
External functions
******************************************************************************/

//! Initialize Configuration parser object.
/*! This function parses command lines arguments 
	and opens configuration file.
  \param    this			pointer to Configuration parser instance
  \param	pChsim			pointer to channel simulator instance
  \param	argc			argument of main
  \param	argv			argument of main
*/
void PARSE_init(	tPARSE_inst * this, tCHSIM_inst		* pChsim,
					int			  argc, char			* argv[] )
{
	const char  * pCfg, * pMsg;

	memset( this, 0, sizeof(*this) ); // not 3-Phase, not binary, no session name, no input file name, no log, no vis

	ASSERT( pChsim );
	this->pChsim = pChsim;

	switch( argc ) {
		case 2:   // configuration file name specified
			pCfg = argv[1];
			pMsg = "log.txt"; // Text message log default name
			break;
		case 3:   // configuration and message log file name specified
			pCfg = argv[1];
			pMsg = argv[2];
			break;
		default: // no arguments or undesired arguments number
			pCfg = "cfg.txt"; // Configuration default name
			pMsg = "log.txt"; // Text message log default name
			break;
	}

	FILEIO_openLog( pMsg ); // first of all open message log file
	this->pFile = fopen( pCfg, "rt" );
	if ( this->pFile ) {
		FILEIO_printf( "Open cfg file %s\n", pCfg );
	} else {
		FILEIO_printf( "Can not open cfg file %s for rt\n", pCfg );
		exit(EXIT_FAILURE);
	}
	
	this->cfg = PARSE_defCfg;
}


//! Parse configuration file.
/*! This function reads and parses configuration file.
	Configuration file may consists from several session.
	Result is returned at each end of session.
    Return NULL if no more session available. 
  \param      this			pointer to Configuration parser instance
  \return					pointer to session configuration, NULL on end of file
*/
tPARSE_cfg * PARSE_parseCfg( tPARSE_inst * this )
{
	char					  aBuf[PARSE_STR_SZ];
	char					* pStr, * ptr;
	const tPARSE_cmdDesc	* pCmd;

	// Parse configuration file string by string
	while ( fgets ( aBuf, PARSE_STR_SZ, this->pFile ) != NULL ) {
		this->strNumb++; // preincrement to count string from 1

		pStr = &aBuf[ strspn( aBuf, pWhite ) ]; // remove first nonwhite characters

		ptr = strchr( pStr, ';' ); // find comments
		if ( ptr )
		  *ptr = '\0'; // remove it

		if ( strlen (pStr) == 0 )
		  continue;  // ignore empty string

		ptr = pStr; 
		while ( *ptr = tolower( *ptr ) ) 
		  ptr++;  // convert to lower case (platform independant)

		ptr = pStr + strcspn( pStr, pWhite ) ;  // next delimiter
		ptr += strspn( ptr, pWhite );           // next token

		// find command in string
		for ( pCmd = PARSE_aCmdDesc; pCmd->pName != NULL; pCmd++ ) {
			size_t len = strlen( pCmd->pName );
			if ( strncmp( pStr, pCmd->pName, len ) == 0 ) // command name found in list of command
				if ( pCmd->fProc( this, ptr, pCmd->val ) ) { // argument are succesfully processed
					if ( pCmd->iEnd ) { // it is end of session
						PARSE_printCfg( &this->cfg );
						return &this->cfg; // it is last command in session - return configuration
					}
					break;
				}
		}
		if( pCmd->pName == NULL ) // command is not found in list
			if( ! CHSIM_parseStr( this->pChsim, pStr )  ) // try to parse it as channel simulator command
				FILEIO_printf( "!!! Error, can not parse string number %d: %s", (int)this->strNumb, aBuf );
	}

	// all command processed
	fclose( this->pFile );
	return (tPARSE_cfg *)NULL;
}


/******************************************************************************
Internal functions
******************************************************************************/

/*! Print out parameters of session, to be started
*/
void PARSE_printCfg( const tPARSE_cfg * pCfg )
{
	uint16		i;
	bool		iSave;

	FILEIO_printf( "New session\n" );  
	FILEIO_printf( "  Log period: %ld ms, duration: %ld %s\n", (long)pCfg->logPerMs,
					(long)pCfg->sesLen, ( ( pCfg->iLenInBits ) ? "bits" : "ms" )  );

	iSave = false;
	for( i = 0; i < 2; i++ )
		iSave |= pCfg->aiSaveSamp[i];
	if( iSave ) {
		FILEIO_printf( "  Save " );
		for( i = 0; i < 2; i++ )
			if( pCfg->aiSaveSamp[i] )
				FILEIO_printf( "TrRc%d ", (int)i );
		FILEIO_printf( "samples\n" );
	}

	iSave = false;
	for( i = 0; i < 4; i++ )
		iSave |= pCfg->aiSaveBits[i];
	if( iSave ) {
		FILEIO_printf( "  Save " );
		for( i = 0; i < 4; i++ )
			if( pCfg->aiSaveBits[i] )
				FILEIO_printf( "%s%d ", ( ( i&0x02 ) ? "Rc" : "Tr" ), (int)(i&0x01) );
		FILEIO_printf( "bits\n" );
	}
}

//! Command processing function - Save TrRc samples
static bool PARSE_procSaveSamp( tPARSE_inst * this, char * pArg, int val )
{
	uint32		  iSave;
	char		* pEnd;

	iSave = strtoul( pArg, &pEnd, 10 ); 
	pEnd += strspn( pEnd, pWhite ); // span white characters after digits
	if( *pEnd != '\0' ) // string is not fully parsed
		return false;
	if( iSave >= 2 ) // only 0 and 1 is allowed
		return false;

	this->cfg.aiSaveSamp[val] = iSave;
	return true;
}

//! Command processing function - Save bits
static bool PARSE_procSaveBits( tPARSE_inst * this, char * pArg, int val )
{
	uint32		  iSave;
	char		* pEnd;

	iSave = strtoul( pArg, &pEnd, 10 ); 
	pEnd += strspn( pEnd, pWhite ); // span white characters after digits
	if( *pEnd != '\0' ) // string is not fully parsed
		return false;
	if( iSave >= 2 ) // only 0 and 1 is allowed
		return false;

	this->cfg.aiSaveBits[val] = iSave;
	return true;
}

//! Command processing function - Set logging period
static bool PARSE_procLogPer( tPARSE_inst * this, char * pArg, int val )
{
	uint32		  logPer;
	char		* pEnd;

	logPer = strtoul( pArg, &pEnd, 10 ); 
	pEnd += strspn( pEnd, pWhite ); // span white characters after digits
	if( *pEnd != '\0' ) // string is not fully parsed
		return false;
	if( ( logPer == 0 ) || ( logPer > 86400000 ) ) // > 1 day
		return false;

	this->cfg.logPerMs = logPer;
	return true;
}

//! Command processing function - Run session
static bool PARSE_procRun( tPARSE_inst * this, char * pArg, int val )
{
	uint32		  sesLen;		// in ms or in bits
	char		* pEnd;

	sesLen = strtoul( pArg, &pEnd, 10 ); 
	pEnd += strspn( pEnd, pWhite ); // span white characters after digits
	if( *pEnd != '\0' ) // string is not fully parsed
		return false;
	if( ( sesLen == 0 ) || ( sesLen > 86400000 ) ) // > 1 day
		return false;

	this->cfg.sesLen = sesLen;
	this->cfg.iLenInBits = (bool)val;
	return true;
}




