/****************************************************************************/
/*! \file		tben.c
	\brief		Phone line channel simulator
	\author		Iliya Voronov, http://powerdsp.narod.ru
	\version	0.0
     
	 Modem test bench.
	 2 modem instances are connected throught channel simulator.
	 BER tester is connected to modems to generate transmit data and
	 check received data. Modem data and samples are saved to files.
	 Test bench is configured with text comands from config file.
	 Test mode, BER and other results are saved to log file.
*/
/*****************************************************************************/

#include <string.h>
#include <math.h>

#include "parse.h"
#include "chsim.h"
#include "fileio.h"
#include "mdm.h"
#include "bert.h"


/******************************************************************************
Internal definitions
******************************************************************************/

//! Modem number in test bench
#define TBEN_MDM_NUM 2


//! Host structure for modem callback
typedef struct tagTBEN_mdmHst {
	uint16					num;	//!< modem number
	struct tagTBEN_inst	  * pTben;	//!< pointer to test bench
} tTBEN_mdmHst;


//! Modem state
typedef struct tagTBEN_mdmStat {
	tMDM_stat				mdm;	//!< modem internal state
	uint32					trByte;	//!< transmitted byte counter
	uint32					rcByte;	//!< received byte counter
} tTBEN_mdmStat;


/******************************************************************************
Test bench instance definition
******************************************************************************/
typedef struct tagTBEN_inst {

	tPARSE_inst		pars;		//!< command file parser
	tCHSIM_inst		chsim;		//!< channel simulator
	tFILEIO_inst	fileio;		//!< fileIO module
	tBERT_inst		aBert[TBEN_MDM_NUM];	//!< BER tester Tr1->Rc0, Tr0-Rc1
	tMDM_inst		aMdm[TBEN_MDM_NUM];		//!< modem
	tTBEN_mdmHst	aHst[TBEN_MDM_NUM];		//!< modem host for callback
	tTBEN_mdmStat	aStat[TBEN_MDM_NUM];	//!< current modem state

	uint32			msCntr;		//!< ms counter
	uint32			bitCntr;	//!< bit counter
	uint32			logCntr;	//!< log counter in ms

} tTBEN_inst;

static tTBEN_inst TBEN_inst;



/******************************************************************************
Forward and inline declarations
******************************************************************************/


//! Call back function for modem
static void TBEN_mdmCb( void * pHst, eMDM_evtType evt, const void * pArg );


//! Log header
static void TBEN_logHdr( tTBEN_inst * this );

//! Log simulation results
static void TBEN_logRes( tTBEN_inst * this );

//! Log footer
static void TBEN_logFtr( tTBEN_inst * this );

/******************************************************************************
External functions
******************************************************************************/
int main( int argc, char * argv[] )
{
	uint16			i;
	tTBEN_inst	  * this = &TBEN_inst; // for uniformity
	tPARSE_cfg	  * pCfg;
	uint32		  * pCmpCntr;

	PARSE_init( &this->pars, &this->chsim, argc, argv );
	CHSIM_init( &this->chsim );
	FILEIO_init( &this->fileio );

	for( i = 0; i < TBEN_MDM_NUM; i++ ) {
		tMDM_evtHndl	mdmEvtHndl;

		BERT_init( &this->aBert[i] );
		this->aHst[i].num   = i;
		this->aHst[i].pTben = this;
		mdmEvtHndl.pHst = &this->aHst[i];
		mdmEvtHndl.fCb  = TBEN_mdmCb;
		MDM_init( &this->aMdm[i], &mdmEvtHndl );
	}

	while( ( pCfg = PARSE_parseCfg( &this->pars ) ) != NULL ) {
		uint16			frSz1;  // frame size for modem 1, TBEN_FR_SZ +- 1
		uint16			ind1;	// index of nex sample to write in aTr1/aRc1 
		fract2q13		aTr0[MDM_FR_SZ], aRc0[MDM_FR_SZ], aTr1[MDM_FR_SZ*2+2], aRc1[MDM_FR_SZ*2+2];
		ind1 = 0;
		memset( aTr0, 0, sizeof(aTr0) );
		memset( aRc0, 0, sizeof(aRc0) );
		memset( aTr1, 0, sizeof(aTr1) );
		memset( aRc1, 0, sizeof(aRc1) );

		// CHSIM configured in PARSE_parseCfg
		FILEIO_config( &this->fileio, pCfg );
		CHSIM_staNew( &this->chsim );

		for( i = 0; i < TBEN_MDM_NUM; i++ ) {
			tMDM_cfg		mdmCfg = MDM_defCfg;

			BERT_setMode( &this->aBert[i], BERT_MD_PR_511 );
//			BERT_setMode( &this->aBert[i], BERT_MD_ALT_11 );
			mdmCfg.iCall  = i & 0x01; // 0 - Answer, 1 - Calling
			MDM_config( &this->aMdm[i], &mdmCfg );
			MDM_ctrl( &this->aMdm[i], MDM_CC_DISCON, NULL );
			MDM_ctrl( &this->aMdm[i], MDM_CC_CONN, NULL );
			memset( &this->aStat[i], 0, sizeof(this->aStat[0]) );
		}

		TBEN_logHdr( this );
		this->msCntr  = 0;
		this->bitCntr = 0;
		this->logCntr = 0;
		pCmpCntr = ( pCfg->iLenInBits ) ? &this->bitCntr : &this->msCntr;

		do {
			MDM_proc( &this->aMdm[0], aTr0, aRc0 );
			FILEIO_wrSamp( &this->fileio, 0, aTr0,	aRc0, MDM_FR_SZ );
			frSz1 = CHSIM_proc(	&this->chsim,	MDM_FR_SZ,
								aTr0,			aRc0,	
								&aTr1[ind1],	&aRc1[ind1] );
			ind1 += frSz1;
			while( ind1 >= MDM_FR_SZ ) {
				MDM_proc( &this->aMdm[1], aTr1, aRc1 );
				FILEIO_wrSamp( &this->fileio, 1, aTr1,	aRc1, MDM_FR_SZ );
				ind1 -= MDM_FR_SZ;
				memcpy( &aTr1[0], &aTr1[MDM_FR_SZ], ind1*sizeof(int16) );
				memcpy( &aRc1[0], &aRc1[MDM_FR_SZ], ind1*sizeof(int16) );
			}
			FILEIO_flush( &this->fileio );

			this->msCntr  += MDM_FR_MS;	// relay on mdm0
			this->bitCntr = this->aStat[0].rcByte * 8;
			this->logCntr += MDM_FR_MS;
			if( this->logCntr >= pCfg->logPerMs ) {
				this->logCntr = 0;
				TBEN_logRes( this );
			}
		} while( *pCmpCntr < pCfg->sesLen );

		TBEN_logFtr( this );
		FILEIO_close( &this->fileio );
	}

}





/******************************************************************************
Internal functions
******************************************************************************/

//! Call back function for modem
void TBEN_mdmCb( void * pHst, eMDM_evtType evt, const void * pArg )
{
	tTBEN_inst	* this;
	uint16		  n;		// modem number
	tMDM_dbuf	* pDbuf;
	ASSERT( pHst );
	ASSERT( evt < MDM_ET_NUM );
	ASSERT( pArg );

	this	= ((tTBEN_mdmHst*)pHst)->pTben;
	n		= ((tTBEN_mdmHst*)pHst)->num;

	switch( evt ) {
		case MDM_ET_CHNG_ST:
			this->aStat[n].mdm = *(tMDM_stat*)pArg;
			break;
		case MDM_ET_PUT_RC:
			pDbuf = (tMDM_dbuf*)pArg;
			BERT_check( &this->aBert[ n ], pDbuf->aBuf, pDbuf->rqSz );
			FILEIO_wrBits( &this->fileio, (ePARSE_bitsFile)(PARSE_BF_RC0 + n), pDbuf->aBuf, pDbuf->rqSz );
			pDbuf->grSz = pDbuf->rqSz;
			this->aStat[n].rcByte += pDbuf->rqSz;
			break;
		case MDM_ET_GET_TR:
			pDbuf = (tMDM_dbuf*)pArg;
			BERT_gen( &this->aBert[ n^1 ], pDbuf->aBuf, pDbuf->rqSz );
			FILEIO_wrBits( &this->fileio, (ePARSE_bitsFile)(PARSE_BF_TR0 + n), pDbuf->aBuf, pDbuf->rqSz );
			pDbuf->grSz = pDbuf->rqSz;
			this->aStat[n^1].trByte += pDbuf->rqSz;
			break;
	}
}



//! Log header
void TBEN_logHdr( tTBEN_inst * this )
{
	FILEIO_printf( "|------|------|------|--------|--------|\n" );
	FILEIO_printf( "| Time |  Tr  |  Rc  |   BER  | estSNR |\n" );
}


//! Log simulation results
void TBEN_logRes( tTBEN_inst * this )
{
	FILEIO_printf( "|%5ld |%5ld |%5ld |%s|%s|\n", (long)this->msCntr, 
		(long)this->aStat[1].trByte*8, (long)this->aStat[0].rcByte*8,
		BERT_getBer( &this->aBert[0] ), "        "						);
}

//! Log footer
void TBEN_logFtr( tTBEN_inst * this )
{
	tBERT_res	aBer[2];
	BERT_getErr( &this->aBert[0], &aBer[0] );
	BERT_getErr( &this->aBert[1], &aBer[1] );

//	TBEN_logRes( this );
	FILEIO_printf( "|------|------|------|--------|--------|\n" );
	FILEIO_printf( "SNR: %.2f dB, ",	20.*log10( 32768. / (double)this->chsim.cfg.awgnAtt ) );
	FILEIO_printf( "BER: %s-",	BERT_getBer( &this->aBert[0] ) );
	FILEIO_printf( "%s\n",		BERT_getBer( &this->aBert[1] ) );
	BERT_printFull2ch(	this->aBert );
	FILEIO_printf( "\n" );
}
